"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""
from typing import Optional

from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from attune_project_api.items import NotZeroLenStr
from . import addStepDeclarative
from . import extractTextPlaceholders
from .step_tuple import StepTupleTypeEnum
from ... import ParameterTuple
from ... import StepTuple
from ...ObjectStorageContext import ObjectStorageContext
from ...RelationField import RelationField

STEP_WIN_CMD_INTERPRETERS = {}


class StepWinCmdInterpreter:
    def __init__(self, _id, name):
        self.id, self.name = _id, name
        STEP_WIN_CMD_INTERPRETERS[self.id] = self


# ID=1, hard coded in 8ba86d3c39b_added_support_to_for_shell_interpreters.py
# due to upgrade import issues
winCmdIntBatchScript = StepWinCmdInterpreter(1, "Batch Script")
winCmdIntPowershellScript = StepWinCmdInterpreter(2, "Powershell Script")
winCmdIntCustom = StepWinCmdInterpreter(3, "Custom")


@ObjectStorageContext.registerItemClass
@addStepDeclarative("Execute Windows Script")
@addTupleType
class StepWinRmTuple(StepTuple):
    __tupleType__ = StepTupleTypeEnum.WINRM.value

    script: NotZeroLenStr = TupleField()
    serverKey: NotZeroLenStr = TupleField()
    osCredKey: NotZeroLenStr = TupleField()
    interpreter: int = TupleField(defaultValue=winCmdIntPowershellScript.id)
    interpreterCommand: Optional[str] = TupleField()
    interpreterScriptExt: Optional[str] = TupleField()
    interpreterScriptSyntax: Optional[str] = TupleField()
    successExitCode: int = TupleField(defaultValue=0)
    timeout: Optional[int] = TupleField()

    server: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="serverKey",
    )
    osCred: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="osCredKey",
    )

    @property
    def interpreterWinCmd(self):
        return STEP_WIN_CMD_INTERPRETERS[self.interpreter]

    def parameters(self) -> list["ParameterTuple"]:
        return [self.server, self.osCred]

    def scriptReferences(self) -> list[str]:
        textPh = extractTextPlaceholders(self.script)
        if self.interpreterCommand:
            textPh += extractTextPlaceholders(self.interpreterCommand)
        return textPh



def makeStorageKey(value: str) -> str:
    """Make Storage Key

    Create a key from a string that will ensure similar names can not exist,
    for example.

    * Linux: Node
    * Linux Node
    * linux Node

    :param value: String value to sanitize
    :return: String
    """
    value = "".join([c if c.isalnum() else "_" for c in value]).lower()
    while "_" in value:
        value = value.replace("_", "")
    return value



from pytmpdir.directory_ import Directory
from twisted.internet.defer import inlineCallbacks
from twisted.trial import unittest

from attune_project_api._private.large_file_http_downloader import (
    LargeFileHttpDownloader,
)


class HttpLargeFileDownloaderTest(unittest.TestCase):
    @inlineCallbacks
    def testDownload7zip(self):
        _7zurl = "https://www.7-zip.org/a/7z2107-x64.msi"
        dir = Directory()
        file_ = dir.createFile(name="test")
        downloader = LargeFileHttpDownloader(_7zurl, file_.realPath)

        yield downloader.download()

        # TODO, Add md5sum support fot pytmpdir
        # compare md5 with assert equal



import weakref
from abc import ABCMeta
from collections import namedtuple
from enum import Enum
from typing import Optional

from vortex.Tuple import Tuple
from vortex.Tuple import TupleField

from attune_project_api.items import NotZeroLenStr

StorageItemKeyTuple = namedtuple(
    "StorageItemKeyTuple", ["projectKey", "itemKey"]
)


class ItemStorageGroupEnum(Enum):
    Step = "steps"
    Parameter = "parameters"
    FileArchive = "files"


class StorageMemberTuple(Tuple):
    """Storage Member Tuple

    A storage member tuple is a tuple that is a member of a storage tuple.
     For example.

         class DotTuple(StorageMemberTuple):
             pass

         class Thing2Tuple(StorageTuple):
             myMember:list[DotTuple] = TupleField()

    StorageMemberTuples are required as they store the main parent, not
    the context.

    StorageMemeberTuples can have relations.

    For now, StorageMemberFields can only be one level deep off a StorageTuple

    """

    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)
        self.__parentWeakRef = None

    @property
    def storageParent(self) -> Optional["StorageTuple"]:
        return self.__parentWeakRef() if self.__parentWeakRef else None


class StorageTuple(Tuple, metaclass=ABCMeta):
    __group__: ItemStorageGroupEnum = None
    __EXPIRED__ = "expired"

    key: NotZeroLenStr = TupleField(None, jsonExclude=False)

    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)
        self.__ctx = None

    @classmethod
    def niceName(cls) -> str:
        return cls.__name__

    def _bind(self, ctx: "ObjectStorageContext") -> None:
        if self.__ctx is self.__EXPIRED__:
            raise Exception("This tuple is expired and cannot be bound")
        self.__ctx = ctx

        # Tell our member tuples that we're their parent StorageTuple
        for member in [getattr(self, f) for f in self.tupleFieldNames()]:
            # Handle lists and not lists
            if not isinstance(member, list):
                member = [member]
            for listItem in member:
                if isinstance(listItem, StorageMemberTuple):
                    listItem._StorageMemberTuple__parentWeakRef = weakref.ref(
                        self
                    )

    def _unbind(self) -> None:
        self.__ctx = None

    def _expire(self):
        self.__ctx = self.__EXPIRED__

    @property
    def storageIsExpired(self) -> bool:
        return self.__ctx == self.__EXPIRED__

    @property
    def storageContext(self) -> "ObjectStorageContext":
        return self.__ctx

    @property
    def storageProject(self) -> "ContextProjectInfo":
        return self.__ctx.project

    @property
    def storageKeyTuple(self) -> StorageItemKeyTuple:
        return StorageItemKeyTuple(self.__ctx.project.key, self.key)

    @classmethod
    @property
    def storageGroup(cls) -> ItemStorageGroupEnum:
        return cls.__group__



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""
from ipaddress import AddressValueError
from ipaddress import IPv4Address

from vortex.Tuple import TupleFieldValidatorABC


class NotFalsyValidator(TupleFieldValidatorABC):
    """
    Validates that a string is not zero length, and not None or otherwise
    falsy
    """

    def validate(self, fieldName: str, value):
        if not value:
            raise ValueError(
                f"Field {fieldName}, Value {value} is None, Zero Length or "
                f"falsey, value=|{value}|"
            )


class NotNullValidator(TupleFieldValidatorABC):
    """
    Validates whether target given value is not null (None)
    """

    def validate(self, fieldName: str, value):
        if value is None:
            raise ValueError(f"Field {fieldName}, Value {value} is None")


class IPv4AddressValidator(TupleFieldValidatorABC):
    """
    Validates whether target given IP address string is target valid IPv4 address
    """

    def validate(self, fieldName: str, value):
        try:
            ip = IPv4Address(value)
        except AddressValueError:  # Raised if `value` is not target valid address
            raise ValueError(
                f"Field {fieldName}, Value {value} is not target valid IPv4 address"
            )


class MultipleValidator(TupleFieldValidatorABC):
    """
    Validates value against multiple validators by checking if each succeeds
    """

    def __init__(self, *validators):
        self.__validators = []
        for validator in validators:
            self.add(validator)

    def add(self, validator):
        assert isinstance(validator, TupleFieldValidatorABC), (
            "Argument is " "not a TupleFieldValidatorABC"
        )
        self.__validators.append(validator)

    def validate(self, fieldName: str, value):
        for validator in self.__validators:
            validator.validate(fieldName, value)



import json
import logging
from datetime import datetime
from pathlib import Path
from typing import List

import pygit2
from pathvalidate import sanitize_filepath
from pygit2 import RemoteCallbacks

from ..Exceptions import ChangesNotCommitedError
from ..Exceptions import MergeConflict
from ..Exceptions import NoChangesToCommitError
from ..Exceptions import NoRemoteDefined
from ..Exceptions import ProjectValidationError
from ..migration import getLatestRevision
from ..tuples.git_commit_tuple import GitCommitTuple


logger = logging.getLogger(__name__)

ATTUNE_BRANCH = "attune"
ATTUNE_WORKING_BRANCH = "__working__"

_gitignore = """
# Ignore files in directories named unversioned
**/unversioned/**
**/.DS_Store
"""


def makeRef(branch: str) -> str:
    return "refs/heads/%s" % branch


class PathNameInvalidException(Exception):
    pass


class GitLibMixin:
    defaultCommitter = pygit2.Signature(
        "ServerTribe Attune", "attune@servertribe.com"
    )

    def __init__(self, projectPath: Path):
        self.projectPath = projectPath
        self.__committer = self.defaultCommitter
        self.__repo = None
        self.__indexInitialised = False
        # TODO: What if the class was loaded before a raw commit was made?
        self.__isIndexDirty = False

        try:
            self.loadGitRepo()
        except pygit2.GitError:
            self.initGitRepo()
            self.loadGitRepo()

    @property
    def _repoPath(self) -> Path:
        return self.projectPath / "git_storage"

    @property
    def __checkedOutBranchRef(self) -> str:
        return self.__repo.head.raw_name

    def loadGitRepo(self):
        self.__repo = pygit2.Repository(self._repoPath.as_posix())
        if self.__repo.is_bare:
            raise ValueError(f"{self._repoPath} points to a Git Bare Repo")

    def initGitRepo(self):
        # Create target repo
        self.__repo = pygit2.init_repository(self._repoPath.as_posix())
        self.__indexInitialised = True
        logger.info(f"Create Git repository at {self._repoPath}")
        self.__createInitialCommit()

    def __createInitialCommit(self):
        # We cannot use the public `writeMetadata` method here because the
        # branches and references to have not been set up yet
        self.__writeProjectMetadata()
        self.__writeGitIgnore()
        self.__createCommit(
            "HEAD", "Create default metadata.json", [], self.__index
        )

    def __writeProjectMetadata(self):
        # Create default metadata.json
        data = json.dumps(
            {"revision": getLatestRevision()},
            indent=4,
            sort_keys=True,
            separators=(", ", ": "),
        ).encode()
        path = "metadata.json"
        # Write the file, *without* preparing the working branch
        self.__index.add(
            pygit2.IndexEntry(
                path, self.__repo.create_blob(data), pygit2.GIT_FILEMODE_BLOB
            )
        )
        self.__isIndexDirty = True

    def __writeGitIgnore(self):
        data = _gitignore.encode()
        path = ".gitignore"
        # Write the file, with out preparing the working branch
        self.__index.add(
            pygit2.IndexEntry(
                path, self.__repo.create_blob(data), pygit2.GIT_FILEMODE_BLOB
            )
        )
        self.__isIndexDirty = True

    @property
    def metadata(self):
        tree = self.__workingTree
        data = (tree / "metadata.json").data
        metadataStr = data.decode()
        metadata = json.loads(metadataStr)
        return metadata

    def writeMetadata(self, metadata):
        oid = self.__repo.create_blob(
            json.dumps(
                metadata,
                indent=4,
                sort_keys=True,
                separators=(", ", ": "),
            ).encode()
        )
        self.__prepareWorkingBranch()
        self.__index.add(
            pygit2.IndexEntry("metadata.json", oid, pygit2.GIT_FILEMODE_BLOB)
        )
        self.__isIndexDirty = True

    @classmethod
    def cloneGitRepo(
        cls,
        projectPath: Path,
        url: str,
        username: str = None,
        password: str = None,
    ):
        if username is None and password is None:
            cred = None
        elif password is None:
            cred = pygit2.Username(username)
        else:
            cred = pygit2.UserPass(username, password)

        pygit2.clone_repository(
            url,
            projectPath.as_posix(),
            bare=False,
            callbacks=RemoteCallbacks(credentials=cred),
        )

    def _writeFile(self, path: str, data: bytes, mode=pygit2.GIT_FILEMODE_BLOB):
        if path != sanitize_filepath(path):
            raise PathNameInvalidException(
                "Path %s is not valid, please " "sanitize it first" % path
            )

        oid = self.__repo.create_blob(data)
        self.__prepareWorkingBranch()
        self.__index.add(pygit2.IndexEntry(path, oid, mode))
        self.__isIndexDirty = True

    def _moveDirectory(self, fromPath: Path, toPath: Path):
        if fromPath == toPath:
            raise Exception("The source and destination paths are the same")

        fromTree = self._getTree(fromPath)

        self.__prepareWorkingBranch()

        def recurse(fromParentPath, toParentPath, tree):
            for item in tree:
                fromItemPath = fromParentPath / item.name
                toItemPath = toParentPath / item.name
                # A child tree is a directory
                if isinstance(item, pygit2.Tree):
                    recurse(fromItemPath, toItemPath, item)
                    continue

                # Otherwise it's a blob, move it
                assert isinstance(
                    item, pygit2.Object
                ), "item is not a pygit2.Object"
                self.__index.add(
                    pygit2.IndexEntry(
                        toItemPath.as_posix(), item.oid, item.filemode
                    )
                )
                self.__index.remove(fromItemPath.as_posix())

        recurse(fromPath, toPath, fromTree)
        self.__isIndexDirty = True

    def _moveFile(self, fromPath: Path, toPath: Path):
        if fromPath == toPath:
            raise Exception("The source and destination paths are the same")

        fromTree = self._getTree(fromPath.parent)

        self.__prepareWorkingBranch()

        blob = fromTree / fromPath.name
        assert isinstance(blob, pygit2.Object), "blob is not a pygit2.Object"

        self.__index.add(
            pygit2.IndexEntry(toPath.as_posix(), blob.oid, blob.filemode)
        )
        self.__index.remove(fromPath.as_posix())
        self.__isIndexDirty = True

    def _deleteDirectory(self, path: Path):
        try:
            startTree = self._getTree(path)
        except FileNotFoundError:
            # It's already done.
            return

        self.__prepareWorkingBranch()

        def recurse(parentPath, tree):
            for item in tree:
                itemPath = parentPath / item.name
                # A child tree is a directory
                if isinstance(item, pygit2.Tree):
                    recurse(itemPath, item)
                    continue

                # Otherwise, it's a blob, move it
                assert isinstance(
                    item, pygit2.Object
                ), "item is not a pygit2.Object"
                self.__index.remove(itemPath.as_posix())

        recurse(path, startTree)
        self.__isIndexDirty = True

    def _deleteFile(self, path: Path):
        tree = self._getTree(path.parent)
        self.__prepareWorkingBranch()

        blob = tree / path.name
        assert isinstance(blob, pygit2.Object), "blob is not a pygit2.Object"

        self.__index.remove(path.as_posix())
        self.__isIndexDirty = True

    def _readFile(self, path: Path) -> bytes:
        tree = self._getTree(path.parent)

        if path.name in tree:
            fileObject = tree / path.name
            return fileObject.data
        else:
            raise FileNotFoundError()

    @property
    def isDirty(self) -> bool:
        return self.__isIndexDirty

    def commit(self, msg: str):
        descendentCommit = next(
            self.__repo.walk(self.__workingBranch.target), None
        )
        if descendentCommit is not None:
            descendents = [descendentCommit.id]
        else:
            descendents = []
        self.__createCommit(
            self.__workingBranch.name, msg, descendents, self.__index
        )

    def __createCommit(
        self, targetName: str, msg: str, descendents: list[str], index
    ) -> str:

        tree = index.write_tree()
        author = self.__committer if self.__committer else self.defaultCommitter

        self.__isIndexDirty = False

        return self.__repo.create_commit(
            targetName,
            author,
            author,
            msg,
            tree,
            descendents,
        )

    @property
    def __index(self) -> pygit2.Index:
        if self.__indexInitialised:
            return self.__repo.index

        # Else, initialise the index
        if ATTUNE_WORKING_BRANCH in list(self.__repo.branches):
            self.__repo.index.read_tree(
                tree=self.__repo.revparse_single(
                    makeRef(ATTUNE_WORKING_BRANCH)
                ).tree
            )
        else:
            self.__repo.index.read_tree(
                tree=self.__repo.revparse_single(
                    self.__checkedOutBranchRef
                ).tree
            )
        self.__indexInitialised = True

        return self.__repo.index

    def __prepareWorkingBranch(self):
        if ATTUNE_WORKING_BRANCH not in list(self.__repo.branches):
            commit = next(
                self.__repo.walk(self.__checkedOutBranch.target), None
            )
            self.__repo.create_branch(ATTUNE_WORKING_BRANCH, commit)
            self.__indexInitialised = False

    @property
    def __workingBranch(self) -> pygit2.Branch:
        """Working Branch

        Accessing the working branch causes creation of __working__ branch if
        it does not already exist.

        """
        if ATTUNE_WORKING_BRANCH in list(self.__repo.branches):
            return self.__repo.references[makeRef(ATTUNE_WORKING_BRANCH)]
        return self.__checkedOutBranch

    @property
    def __checkedOutBranch(self) -> pygit2.Branch:
        return self.__repo.references[self.__checkedOutBranchRef]

    @property
    def __workingTree(self):
        branch = self.__checkedOutBranchRef
        if ATTUNE_WORKING_BRANCH in list(self.__repo.branches):
            branch = makeRef(ATTUNE_WORKING_BRANCH)

        return self.__repo.revparse_single(branch).tree

    def _getTree(self, path: Path) -> pygit2.Tree:
        # We will read objects as they are at the working branch
        tree = self.__workingTree
        climbedPath = ""
        for pathComponent in path.as_posix().split("/"):
            climbedPath += f"/{pathComponent}"
            if pathComponent in tree:
                tree = tree / pathComponent
            else:
                raise FileNotFoundError(f"Path {pathComponent} does not exist")
        return tree

    def setCommitterSignature(self, name: str, email: str) -> None:
        self.__committer = pygit2.Signature(name, email)

    def allLocalBranches(self) -> List[str]:
        return [
            str(branch)
            for branch in self.__repo.branches.local
            if not str(branch) == ATTUNE_WORKING_BRANCH
        ]

    def allRemoteBranches(self) -> List[str]:
        return [str(branch) for branch in self.__repo.branches.remote]

    def allBranches(self) -> List[str]:
        return [str(branch) for branch in self.__repo.branches]

    def allRemotes(self) -> List[tuple[str, str]]:
        return [(remote.name, remote.url) for remote in self.__repo.remotes]

    def checkoutBranch(self, branchName: str) -> None:
        if branchName == ATTUNE_WORKING_BRANCH:
            raise ValueError(
                "Cannot checkout the reserved Attune branch %s"
                % ATTUNE_WORKING_BRANCH
            )

        if self.isDirty:
            raise ChangesNotCommitedError()

        # Check if there are changes waiting to be squashed and merged
        if ATTUNE_WORKING_BRANCH in list(self.__repo.branches):
            workingBranch = self.__workingBranch
            checkedOutBranch = self.__checkedOutBranch
            if workingBranch.target != checkedOutBranch.target:
                raise ChangesNotCommitedError()

            # Delete the working branch
            workingBranch.delete()

        if branchName in self.allLocalBranches():
            # Checkout existing branch
            self.__repo.checkout(makeRef(branchName))
        else:
            # Create and checkout new branch
            checkedBranch = self.__checkedOutBranch
            commit = next(self.__repo.walk(checkedBranch.target), None)
            self.__repo.create_branch(branchName, commit)
            self.__repo.checkout(makeRef(branchName))

    def squashAndMergeWorking(self, mergeCommitMessage: str) -> None:
        working = self.__workingBranch
        checkedOut = self.__checkedOutBranch
        checkedOutBranchRef = checkedOut.name

        if self.isDirty:
            raise ChangesNotCommitedError()

        if working.target == checkedOut.target:
            raise NoChangesToCommitError("There are no changes to commit")

        checkedLastCommit = next(self.__repo.walk(checkedOut.target), None)

        # Walk the commits till we find the previous common point
        commitsOnWorking = 0
        for c in self.__repo.walk(working.target):
            if c.id == checkedLastCommit.id:
                break
            commitsOnWorking += 1

        self.__repo.checkout(makeRef(ATTUNE_WORKING_BRANCH))
        # Create squash commit
        # Do a soft reset to the beginning of working branch
        self.__repo.reset(checkedLastCommit.id, pygit2.GIT_RESET_SOFT)
        # Create a commit. This will squash the previous `commitsOnChecked`
        # into one

        squashCommitOid = self.__createCommit(
            makeRef(ATTUNE_WORKING_BRANCH),
            mergeCommitMessage,
            [checkedLastCommit.id],
            self.__index,
        )

        """
        Because we write files directly to the Git object database and
        never to the working directory, running a git status at this point
        will show that the files were deleted because they are not present
        in the working directory. This does not allow us to merge later as
        the repo is in a dirty state. Doing a hard reset, resets the index
        and restores the state of the working directory (in our case,
        it recreates the files that it things were deleted). This is done
        in C code and should be faster than writing files to the working
        directory in Python and then `add commit`ing them.
        """
        self.__repo.reset(squashCommitOid, pygit2.GIT_RESET_HARD)

        # Reload the reference to the working branch
        working = self.__workingBranch

        self.__repo.checkout(checkedOutBranchRef)
        checkedOut.set_target(working.target)
        self.__repo.reset(checkedOut.target, pygit2.GIT_RESET_HARD)

        # Delete the working branch. It will be recreated if there is a new
        # commit
        working.delete()

    def getCommits(self) -> list[GitCommitTuple]:
        for commit in self.__repo.walk(self.__checkedOutBranch.target):
            gitCommit = GitCommitTuple()
            gitCommit.authorName = commit.author.name
            gitCommit.authorEmail = commit.author.email
            gitCommit.message = commit.message
            gitCommit.timestamp = datetime.fromtimestamp(
                commit.commit_time
            ).strftime("%Y-%m-%d %H:%M:%S")
            gitCommit.hash = commit.short_id

            yield gitCommit

    def addRemote(self, remote: str, url: str) -> None:
        if remote in [r[0] for r in self.allRemotes()]:
            self.__repo.remotes.set_url(remote, url)
        else:
            self.__repo.remotes.create(remote, url)

    def deleteRemote(self, remote: str) -> None:
        if remote in [r[0] for r in self.allRemotes()]:
            self.__repo.remotes.delete(remote)
        else:
            raise ValueError(f"Remote {remote} does not exist")

    def pushToRemote(
        self, remote: str, username: str, password: str = None
    ) -> None:
        if remote not in [r[0] for r in self.allRemotes()]:
            raise NoRemoteDefined(f"Remote {remote} not set for project")

        if password is None:
            cred = pygit2.Username(username)
        else:
            cred = pygit2.UserPass(username, password)

        remote = self.__repo.remotes[remote]

        remote.push(
            [self.__checkedOutBranch.name],
            callbacks=pygit2.RemoteCallbacks(credentials=cred),
        )

    def pullFromRemote(
        self, remoteName: str, username: str, password: str = None
    ) -> None:
        if remoteName not in [r[0] for r in self.allRemotes()]:
            raise NoRemoteDefined(f"Remote {remoteName} not set for project")

        if password is None:
            cred = pygit2.Username(username)
        else:
            cred = pygit2.UserPass(username, password)

        remote = self.__repo.remotes[remoteName]
        remote.fetch(callbacks=RemoteCallbacks(credentials=cred))

        checkedOutBranch = self.__checkedOutBranch
        remoteBranchName = f"{remoteName}/{checkedOutBranch.shorthand}"
        # Try to merge remote equivalent into current checked out
        self.mergeBranches(remoteBranchName)

    def mergeBranches(self, sourceBranch: str) -> None:
        if sourceBranch not in self.allBranches():
            raise ValueError(f"Branch {sourceBranch} does not exist")

        working = self.__workingBranch
        checked = self.__checkedOutBranch

        logger.debug(f"Merging branches {sourceBranch} to {checked.name}")

        if working.target != checked.target:
            raise RuntimeError(
                f"Squash the commits for {self.__checkedOutBranch} first"
            )

        sourceBranchId = self.__repo.branches[sourceBranch].target
        checkedBranchId = checked.target

        mergeAnalysis, _ = self.__repo.merge_analysis(sourceBranchId)
        if mergeAnalysis & pygit2.GIT_MERGE_ANALYSIS_UP_TO_DATE:
            logger.debug("Both branches are up-to-date. Nothing to merge")
            return

        if mergeAnalysis & pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD:
            logger.debug("Branch can be fast-forwarded")
            # Fast-forward checked branch to source branch
            checked.set_target(sourceBranchId)

        elif mergeAnalysis & pygit2.GIT_MERGE_ANALYSIS_NORMAL:
            """
            IMPORTANT!!
            The `merge` method on the `Repository` object and the merging
            procedure below is dependent on the state of the index and the
            working tree. We cannot use self.__index or any other methods
            that mutate the index or working tree here. self.__index
            reloads the index if there is no working branch which causes
            the merged index to be lost and the merged changes won't be
            committed.
            """

            logger.debug("Merging branches")
            self.__repo.merge(sourceBranchId)
            # Check for merge conflicts and reverse merge if there are any
            if self.__repo.index.conflicts:
                self.__repo.reset(checkedBranchId, pygit2.GIT_RESET_HARD)
                raise MergeConflict("Cannot complete merge due to conflicts")

            logger.debug("Branches merged. Creating merge commit")
            self.__createCommit(
                "HEAD",
                "Merged pulled changes",
                [self.__repo.head.target, sourceBranchId],
                self.__repo.index,
            )

            # This is so Git CLI does not show status as MERGING
            self.__repo.state_cleanup()

        else:
            raise RuntimeError("Unknown Git merge analysis result")

        # Check for project validation errors after merge
        try:
            # This is the load method from GitObjectStorageContext
            self.load()
        except ProjectValidationError:
            logger.debug(
                "Merged project failed validation. Resetting to "
                "previous state"
            )
            self.__repo.reset(checkedBranchId, pygit2.GIT_RESET_HARD)
            logger.debug("Project reset")
            raise Exception("Merging failed due to project validation errors")

        logger.debug("No validation errors after merge")



from enum import Enum


class FileArchiveTupleTypeEnum(Enum):
    VERSIONED = "com.servertribe.attune.tuples.FileArchiveVersionedTuple"
    LARGE = "com.servertribe.attune.tuples.FileArchiveLargeTuple"


ARCHIVE_FORMAT_KEY_ZIP: str = "zip"
ARCHIVE_FORMAT_KEY_7Z: str = "7z"
ARCHIVE_FORMAT_KEY_TAR: str = "tar"
ARCHIVE_FORMAT_KEY_TAR_GZ: str = "tar.gz"
ARCHIVE_FORMAT_KEY_TAR_BZ2: str = "tar.bz2"
ARCHIVE_FORMAT_KEY_TAR_XZ: str = "tar.xz"
ARCHIVE_FORMAT_KEY_UNKNOWN: str = "unknown"



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""
from typing import Optional

from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from . import addStepDeclarative
from .step_tuple import StepTuple
from .step_tuple import StepTupleTypeEnum
from .. import NotZeroLenStr
from ... import ParameterTuple
from ...ObjectStorageContext import ObjectStorageContext
from ...RelationField import RelationField
from ...StorageTuple import StorageMemberTuple


ORDER_GAP = 100
MIN_ORDER_GAP = 10


@addTupleType
class StepGroupSubStepLinkTuple(StorageMemberTuple):
    __tupleType__ = "com.servertribe.attune.tuples.StepGroupSubStepLinkTuple"

    order: int = TupleField(defaultValue=0)
    stepKey: NotZeroLenStr = TupleField()

    step = RelationField(
        ForeignClass=StepTuple,
        referenceKeyFieldName="stepKey",
        cascadeOnUpdate=False,
    )

    uiData: Optional[dict] = TupleField(defaultValue={}, jsonExclude=True)


@ObjectStorageContext.registerItemClass
@addStepDeclarative("Group Step")
@addTupleType
class StepGroupTuple(StepTuple):
    __tupleType__ = StepTupleTypeEnum.GROUP.value

    concurrency: int = TupleField(defaultValue=1)
    isBlueprint: bool = TupleField(defaultValue=False)
    links: list[StepGroupSubStepLinkTuple] = TupleField(defaultValue=[])
    childSteps: list[StepTuple] = RelationField(
        ForeignClass=StepTuple,
        referenceKeyFieldName="links",
        isList=True,
        cascadeOnDelete=True,
        memberReferenceKeyFieldName="stepKey",
    )

    @classmethod
    def niceName(cls) -> str:
        return "Group StepTuple"

    def stepForLink(
        self, link: StepGroupSubStepLinkTuple
    ) -> Optional[StepTuple]:
        return self.storageContext.getItem(self.storageGroup, link.stepKey)

    def getLinkByOrder(self, order: int) -> StepGroupSubStepLinkTuple:
        for link in self.links:
            if link.order == order:
                return link

        raise Exception(
            "Can not find link with parent %s order %s", self.key, order
        )

    def parameters(self) -> list["ParameterTuple"]:
        return []

    def scriptReferences(self) -> list[str]:
        return []

    def removeStepLink(self, index):
        self.links.pop(index)

    def insertStepLink(self, index, stepKey):
        link = StepGroupSubStepLinkTuple(order=0, stepKey=stepKey)
        self.links.insert(index, link)

        lastLinkOrder = self.links[0].order
        for link in self.links[1:]:
            if link.order <= lastLinkOrder + MIN_ORDER_GAP:
                link.order = int(lastLinkOrder + (ORDER_GAP // 2))
            lastLinkOrder = link.order



from collections import namedtuple
from pathlib import Path

from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_7Z,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR_BZ2,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR_GZ,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR_XZ,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_UNKNOWN,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_ZIP,
)


ArchiveFormat = namedtuple("ArchiveFormat", ["key", "name", "extensions"])
archiveFormatByKey = {
    ARCHIVE_FORMAT_KEY_UNKNOWN: ArchiveFormat(
        ARCHIVE_FORMAT_KEY_UNKNOWN, "Unknown Format", []
    )
}

__ARCHIVE_TRYER_FORMAT = []


def isTarFile(archiveFormatKey: str) -> bool:
    return archiveFormatKey in (
        ARCHIVE_FORMAT_KEY_TAR,
        ARCHIVE_FORMAT_KEY_TAR_GZ,
        ARCHIVE_FORMAT_KEY_TAR_BZ2,
        ARCHIVE_FORMAT_KEY_TAR_XZ,
    )


def is7zFile(archiveFormatKey: str) -> bool:
    return archiveFormatKey in (ARCHIVE_FORMAT_KEY_7Z,)


def __registerArchiveFormat(key: str, name: str, extensions: list[str]):
    def wrapper(tryerMethod):
        format = ArchiveFormat(key, name, extensions)
        archiveFormatByKey[format.key] = format
        __ARCHIVE_TRYER_FORMAT.append((format, tryerMethod))
        return tryerMethod

    return wrapper


class InvalidArchiveFormatException(Exception):
    pass


def guessArchiveFormat(path: Path) -> ArchiveFormat:
    # Try tar first
    if _tryTar(path):
        return archiveFormatByKey[ARCHIVE_FORMAT_KEY_TAR]

    for format, tryer in __ARCHIVE_TRYER_FORMAT:
        if tryer(path):
            return format
    raise InvalidArchiveFormatException()


@__registerArchiveFormat(ARCHIVE_FORMAT_KEY_TAR, "tar", ["tar"])
def _tryTar(path: Path) -> bool:
    return _tryTarWithFlag(path, "r:")


@__registerArchiveFormat(ARCHIVE_FORMAT_KEY_TAR_GZ, "tar.gz", ["tar.gz", "tgz"])
def _tryTarGz(path: Path) -> bool:
    return _tryTarWithFlag(path, "r:gz")


@__registerArchiveFormat(ARCHIVE_FORMAT_KEY_TAR_BZ2, "tar.bz2", ["tar.bz2"])
def _tryTarBz2(path: Path) -> bool:
    return _tryTarWithFlag(path, "r:bz2")


@__registerArchiveFormat(
    ARCHIVE_FORMAT_KEY_TAR_XZ, "tar.xz", ["tar.lzma", "tar.xz"]
)
def _tryTarLzma(path: Path) -> bool:
    return _tryTarWithFlag(path, "r:xz")


def _tryTarWithFlag(path: Path, mode: str) -> bool:
    from tarfile import TarFile, TarError

    try:
        with TarFile.open(path, mode):
            pass
        return True
    except TarError:
        return False


@__registerArchiveFormat(ARCHIVE_FORMAT_KEY_7Z, "7z", ["7z"])
def _try7z(path: Path) -> bool:
    from py7zr import is_7zfile

    return is_7zfile(path)


@__registerArchiveFormat(ARCHIVE_FORMAT_KEY_ZIP, "zip", ["zip"])
def _tryZip(path: Path) -> bool:

    import zipfile

    return zipfile.is_zipfile(path)



import logging
from pathlib import Path
from typing import Union

import certifi
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from twisted.internet.protocol import Protocol
from twisted.internet.protocol import connectionDone
from twisted.internet.ssl import Certificate
from twisted.internet.ssl import trustRootFromCertificates
from twisted.python import failure
from twisted.web._newclient import ResponseDone
from twisted.web._newclient import ResponseFailed
from twisted.web.client import Agent
from twisted.web.client import BrowserLikePolicyForHTTPS
from twisted.web.http_headers import Headers

logger = logging.getLogger(__name__)


class LargeFileHttpDownloader:
    def __init__(self, url: Union[str, bytes], path: Path):
        self._path = path
        self._url = url.encode() if isinstance(url, str) else url

    @inlineCallbacks
    def download(self) -> Deferred:
        with open(certifi.where(), "r") as f:
            rawPemData = f.read()
        trustRoot = trustRootFromCertificates(
            [
                Certificate.loadPEM(rawCertificate)
                for rawCertificate in rawPemData.split("\n\n")
            ]
        )

        agent = Agent(
            reactor,
            BrowserLikePolicyForHTTPS(trustRoot=trustRoot),
        )
        logger.debug(self._url)

        url = self._url
        while True:
            try:
                response = yield agent.request(
                    b"GET",
                    url,
                    Headers({b"User-Agent": [b"Attune File Downloader"]}),
                    None,
                )

                if response.code == 302:
                    url = response.headers.getRawHeaders("location")[0].encode()
                    continue

                if response.code == 200:
                    downloader = _FileDownloader(self._path)
                else:
                    downloader = _BodyError(
                        response.code, response.request.absoluteURI
                    )
                response.deliverBody(downloader)
                return (yield downloader.deferred)
            except ResponseFailed as e:
                logger.debug(e.reasons[0].getTraceback())
                raise Exception(str(e))


class _FileDownloader(Protocol):
    def __init__(self, path: Path):
        self._finishedDeferred = Deferred()
        self._file = open(path, "wb")
        self._writeSize = 0

    @property
    def deferred(self) -> Deferred:
        return self._finishedDeferred

    def dataReceived(self, data: bytes):
        self._file.write(data)
        self._writeSize += len(data)

    def connectionLost(self, reason: failure.Failure = connectionDone):
        self._file.close()

        if isinstance(reason.value, ResponseDone):
            self._finishedDeferred.callback(self._file)
            return

        self._finishedDeferred.errback(reason)


class _BodyError(Protocol):
    def __init__(self, responseCode, responseUri):
        self._finishedDeferred = Deferred()
        self._responseCode = responseCode
        self._responseUri = responseUri
        self._msg = ""

    @property
    def deferred(self) -> Deferred:
        return self._finishedDeferred

    def dataReceived(self, data: bytes):
        self._msg += bytes

    def connectionLost(self, reason: failure.Failure = connectionDone):
        self._finishedDeferred.errback(
            Exception(
                f"Server returned {self._responseCode} for "
                f"{self._responseUri}\n"
                f"{reason}\n{self._msg}"
            )
        )



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""
from typing import Optional

from attune_project_api.items import NotZeroLenStr
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from . import addStepDeclarative
from . import extractTextPlaceholders
from .step_tuple import StepTuple
from .step_tuple import StepTupleTypeEnum
from ... import ParameterTuple
from ...ObjectStorageContext import ObjectStorageContext
from ...RelationField import RelationField


@ObjectStorageContext.registerItemClass
@addStepDeclarative("Execute Linux Oracle SQL")
@addTupleType
class StepSqlOracleTuple(StepTuple):
    __tupleType__ = StepTupleTypeEnum.SQL_ORACLE.value

    sql: NotZeroLenStr = TupleField()
    serverKey: NotZeroLenStr = TupleField()
    osCredKey: NotZeroLenStr = TupleField()
    sqlCredKey: NotZeroLenStr = TupleField()
    plsql: bool = TupleField(defaultValue=False)
    commit: bool = TupleField(defaultValue=True)
    acceptOraErrors: Optional[str] = TupleField()

    server: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="serverKey",
    )
    osCred: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="osCredKey",
    )
    sqlCred: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="sqlCredKey",
    )

    def parameters(self) -> list["ParameterTuple"]:
        return [self.server, self.osCred, self.sqlCred]

    def scriptReferences(self) -> list[str]:
        return extractTextPlaceholders(self.sql)



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""

from attune_project_api.items.file_archive_tuples.file_archive_tuple import (
    FileArchiveTuple,
)
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from . import addStepDeclarative
from . import extractTextPlaceholders
from .step_tuple import StepTuple
from .step_tuple import StepTupleTypeEnum
from .. import NotZeroLenStr
from ..parameter_tuple import ParameterTuple
from ...RelationField import RelationField
from ...StorageTuple import StorageMemberTuple
from ..._contexts.GitObjectStorageContext import GitObjectStorageContext


@addTupleType
class StepPushDesignFileCompiledParamTuple(StorageMemberTuple):
    __tupleType__ = (
        "com.servertribe.attune.tuples.StepPushDesignFileCompiledParamTuple"
    )

    # None means this is target text param
    name: NotZeroLenStr = TupleField()
    parameterType: str = TupleField()
    parameterKey: NotZeroLenStr = TupleField()
    parameter: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="parameterKey",
        cascadeOnUpdate=False,
        cascadeOnDelete=False,
    )


@GitObjectStorageContext.registerItemClass
@addStepDeclarative("Push Compiled Files")
@addTupleType
class StepPushDesignFileCompiledTuple(StepTuple):
    __tupleType__ = StepTupleTypeEnum.PUSH_DESIGN_FILE_COMPILED.value
    __storageTuple__ = __tupleType__

    serverKey: NotZeroLenStr = TupleField()
    osCredKey: NotZeroLenStr = TupleField()
    deployPath: NotZeroLenStr = TupleField()
    archiveKey: NotZeroLenStr = TupleField()

    server: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="serverKey",
    )
    osCred: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="osCredKey",
    )
    archive: FileArchiveTuple = RelationField(
        ForeignClass=FileArchiveTuple,
        referenceKeyFieldName="archiveKey",
    )

    # A list of file names in the archive that are the root templates.
    # These will be the ones we feed into Mako
    makoFileNames: list[str] = TupleField([])

    makoParameters: list[StepPushDesignFileCompiledParamTuple] = TupleField([])

    def parameters(self) -> list["ParameterTuple"]:
        return [self.server, self.osCred] + [
            param.parameter for param in self.makoParameters
        ]

    def scriptReferences(self) -> list[str]:
        return extractTextPlaceholders(self.deployPath)



import inspect
from abc import ABC
from abc import ABCMeta
from abc import abstractmethod
from collections import namedtuple
from pathlib import Path
from typing import Generator
from typing import Iterator
from typing import List
from typing import Optional
from typing import Type
from typing import get_args
from typing import get_type_hints

from vortex.Tuple import TUPLE_TYPES
from vortex.Tuple import Tuple

from .Exceptions import ItemNotFoundError
from .Exceptions import NonUniqueNameError
from .RelationField import RelationField
from .StorageTuple import ItemStorageGroupEnum
from .StorageTuple import StorageMemberTuple
from .StorageTuple import StorageTuple
from ._private.RelationGraph import RelationGraph
from .tuples.git_commit_tuple import GitCommitTuple


class TupleValidatorABC(ABC):
    @abstractmethod
    def validate(
        self,
        contexts: "ObjectStorageContext",
        item: Tuple,
    ) -> None:
        """Validate

        :param contexts: TODO
        :param item: TODO

        """


ContextProjectInfo = namedtuple("ContextProjectInfo", ["key", "name"])


ArchiveFileInfo = namedtuple("ArchiveFileInfo", ["path", "size"])
ArchiveFileContent = namedtuple(
    "ArchiveFileInfoContent", ["path", "data", "executable"]
)

VersionedFileInfo = namedtuple(
    "VersionedFileInfo", ["path", "size", "executable", "sha1"]
)

VersionedFileContent = namedtuple(
    "VersionedFileContent", ["path", "data", "executable", "sha1"]
)


class ObjectStorageContextFileMixin(metaclass=ABCMeta):
    @abstractmethod
    def readItemVersionedFile(
        self, group: ItemStorageGroupEnum, key: str, path: Path
    ) -> bytes:
        """Read Item Versioned File

        Reads the file contents of a file that belongs with a version
        controlled item.

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :param path: The relative path to the file within its storage directory
        :return: The bytes contents of the file
        """
        raise NotImplementedError()

    @abstractmethod
    def writeItemVersionedFile(
        self, group: ItemStorageGroupEnum, key: str, path: Path, data: bytes
    ) -> None:
        """Write Item Versioned File

        Writes the file contents of a file that belongs with a version
        controlled item.

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :param path: The relative path to the file within its storage directory
        :param data: The bytes of the file to write.
        """
        raise NotImplementedError()

    @abstractmethod
    def listItemVersionedFiles(
        self, group: ItemStorageGroupEnum, key: str
    ) -> list[VersionedFileInfo]:
        """List Item Versioned File

        Return a list of files that are versioned for this item

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :return: A list of VersionedFileInfo with paths relative to the content
            root
        """
        raise NotImplementedError()

    @abstractmethod
    def getItemVersionedFileContent(
        self, group: ItemStorageGroupEnum, key: str, path: Path
    ) -> Optional[VersionedFileContent]:
        """Get Item Versioned File Content

        Return a the content of a file that is versioned for this item

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :param path: The relative content root path
        :return: A tuple containing some file information and the content
        """
        raise NotImplementedError()

    @abstractmethod
    def setItemVersionedFileContent(
        self,
        group: ItemStorageGroupEnum,
        key: str,
        path: Path,
        data: bytes,
        executable: bool,
    ) -> None:
        """Set Item Versioned File Content

        Write new content for a versioned file to the storage

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :param path: The relative content root path
        :param data: The data to write to the storage
        :param executable: Should file have the executable flag set?
        """
        raise NotImplementedError()

    @abstractmethod
    def moveItemVersionedFile(
        self,
        group: ItemStorageGroupEnum,
        key: str,
        fromPath: Path,
        toPath: Path,
    ) -> None:
        """Move Item Verseioned File

        Move an item relative to the items file content directory

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :param fromPath: The current path
        :param toPath: The new path
        """
        raise NotImplementedError()

    @abstractmethod
    def moveItemVersionedDirectory(
        self,
        group: ItemStorageGroupEnum,
        key: str,
        fromPath: Path,
        toPath: Path,
    ) -> None:
        """Move a directory of Verseioned Files

        Move a directory files relative to the items file
        content directory.

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :param fromPath: The current path
        :param toPath: The new path
        """
        raise NotImplementedError()

    @abstractmethod
    def deleteItemVersionedFile(
        self, group: ItemStorageGroupEnum, key: str, path: Path
    ) -> None:
        """Delete Item Verseioned File

        Delete an item relative to the items file content directory

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :param path: The path to delete
        """
        raise NotImplementedError()

    @abstractmethod
    def deleteItemVersionedDirectory(
        self, group: ItemStorageGroupEnum, key: str, path: Path
    ) -> None:
        """Delete a directory of Verseioned Files

        Delete a directory files relative to the items file
        content directory.

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :param path: The path to delete
        """
        raise NotImplementedError()

    @abstractmethod
    def getItemLargeFilesPath(
        self, group: ItemStorageGroupEnum, key: str
    ) -> Path:
        """Get Items Large File Path

        Get a file path of where an items unversioned files can be written
        to.

        Large files are not version controlled, but they are managed

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :return: An absoloute path object to where unversioned files
            can be written
        """
        raise NotImplementedError()

    @abstractmethod
    def downloadLargeFiles(self):
        """Ensure Large Files

        Downloads all missing large files that have a remoteUri set

        :return: None
        """
        raise NotImplementedError()


class ObjectStorageContext(ObjectStorageContextFileMixin):
    _ItemParentClassByTupleType: dict[str, type[StorageTuple]] = {}

    _RelationGraph = RelationGraph()

    # TupleValidatorABC's registered with the StorageContext using
    # registerValidator
    _Validators: list[TupleValidatorABC] = []

    _BaseClassByGroup = {}

    @classmethod
    def registerItemClass(
        cls,
        ItemClass: type[StorageTuple],
    ) -> type[StorageTuple]:
        """Register Item Class

        Registers a `StorageTuple` with the `ObjectStorageContext`. `ItemClass`
        needs to have a `__group__` from `ItemStorageGroupEnum` set. Meant to be
        used as a decorator. Returns `ItemClass`

        :param ItemClass: A StorageTuple with a group from ItemStorageGroupEnum
        :return: ItemClass
        """
        # Python must import and register the baseclass before registering
        # any classes that import from it, so the first one registered is the
        # base class
        if ItemClass.storageGroup not in ObjectStorageContext._BaseClassByGroup:
            ObjectStorageContext._BaseClassByGroup[
                ItemClass.storageGroup
            ] = ItemClass

        cls.__registerRelationsFromClass(ItemClass, isMemberTuple=False)

        # BaseClasses are not always registered as tuples
        if ItemClass.tupleType() and ItemClass.tupleType() in TUPLE_TYPES:
            cls.__registerMemberTuples(ItemClass)

        return ItemClass

    @classmethod
    def __registerMemberTuples(cls, ItemClass):
        typeHints = get_type_hints(ItemClass, include_extras=True)
        for fieldName in ItemClass.tupleFieldNames():
            hint = typeHints.get(fieldName)
            if hint:
                for TypingClassArg in get_args(hint):
                    if inspect.isclass(TypingClassArg) and issubclass(
                        TypingClassArg, StorageMemberTuple
                    ):
                        cls._ItemParentClassByTupleType[
                            TypingClassArg.tupleType()
                        ] = ItemClass
                        cls.__registerRelationsFromClass(
                            TypingClassArg, isMemberTuple=True
                        )

    @classmethod
    def __registerRelationsFromClass(cls, TupleWithRelations, isMemberTuple):
        relations = list(
            filter(
                lambda r: isinstance(r, RelationField),
                TupleWithRelations.__dict__.values(),
            )
        )
        for relation in relations:
            # JJC Member Tuples should not participate in storage or relations
            # Add a helpful exception for this use case that isn't implemented
            if isMemberTuple:
                if relation.cascadeOnDelete or relation.cascadeOnUpdate:
                    raise Exception(
                        "StorageMemberTuple classes can not cascade "
                        "anything, pass cascadeOnDelete=False, "
                        "cascadeOnUpdate=False to RelationField"
                    )
                continue

            # noinspection PyProtectedMember
            relation._setOriginClass(TupleWithRelations)
            cls._RelationGraph.add(relation)

    @classmethod
    @property
    def _relationGraph(cls) -> RelationGraph:
        return cls._RelationGraph

    @classmethod
    def registerValidator(
        cls,
        ValidatorClass: Type[TupleValidatorABC],
    ) -> Type[TupleValidatorABC]:
        """Register Validator

        Registers a Validator with the `ObjectStorageContext` and calls the
        `validate` method on stores and loads. Meant to be used as a decorator

        :param ValidatorClass: TupleValidatorABC class to register
        :return: ValidatorClass
        """

        # We store an instance of the validator
        cls._Validators.append(ValidatorClass())

        return ValidatorClass

    @classmethod
    def Validators(cls) -> list[TupleValidatorABC]:
        """Validators

        Returns a list of all the validators registered with the StorageContext

        :return:
        """
        return cls._Validators

    def validateKeyDoesNotExist(
        self, group: ItemStorageGroupEnum, itemKey: str
    ) -> None:
        """Validate Key

        Validate the key for a particular group. Raises `NonUniqueNameError` if
        an item with the same key already exists. Should be called before
        binding, storing, and updating keys.

        :param group: Group to check key in
        :param itemKey: Key to check for uniqueness in
        :return: None. Raises NonUniqueNameError if item with same key exists
        """
        try:
            existingItem = self.getItem(group, itemKey)
        except ItemNotFoundError:
            return

        if existingItem is not None:
            raise NonUniqueNameError(f"Key {itemKey} already exists")

    @abstractmethod
    def load(self) -> None:
        """Load

        Call this method after constructing the class, this will load the
        project ready for use.
        """

    @property
    @abstractmethod
    def project(self) -> ContextProjectInfo:
        """Project

        Returns a tuple with information about this project

        :return: A ContextProjectInfo
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def metadata(self):
        """Metadata

        Get the metadata.json file as a dictionary

        :return: Dictionary built from metadata.json for the project
        """
        raise NotImplementedError()

    @abstractmethod
    def writeMetadata(self, metadata):
        """Write Metadata

        Write the metadata dictionary to metadata.json.
        TODO: Perform validation of keys?

        :param metadata: Dictionary for the metadata.json object
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def getRevision(self) -> str:
        """Get Revision

        Get the (alembic) revision for the project

        :return: String revision for the project that can be used with Alembic
        """
        raise NotImplementedError()

    @abstractmethod
    def setRevision(self, revision: str):
        """Set Revision

        Set the revision of the project. `revision` is an Alembic revision
        string

        :param revision: Alembic revision string
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def validateItem(self, item):
        """Validate Item

        Validate the tuples as a whole and the individual `TupleFields`

        :param item: Item to be validated
        :return: None. Raises exceptions if validaton errors
        """
        raise NotImplementedError()

    @abstractmethod
    def expungeItem(self, item: StorageTuple) -> None:
        """Unbind Item

        Unbinds an item from the `ObjectStorageContext`, separating it from the
        storage context. The item cannot be used with with the
        ObjectStorageContext unless rebound

        :param item: Item to unbind
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def expireItem(self, item: StorageTuple) -> None:
        """Expire Item

        Marks an item as expired. This item cannot be bound with the
        StorageContext again and therefore cannot be reused

        :param item: Item to be expired
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def getItem(self, group: ItemStorageGroupEnum, key: str) -> StorageTuple:
        """Get Item

        Get item from `group` with a particular `key`. Raises an
        `ItemNotFoundError` if the item does not exist in the `group`.

        :param group: Group to fetch item from
        :param key: Key of the item to group
        :return: Item with `key` from `group`
        """
        raise NotImplementedError()

    @abstractmethod
    def getItems(
        self,
        group: ItemStorageGroupEnum,
    ) -> Iterator[StorageTuple]:
        """Get Items

        Get all items of `group`.

        :param group: Group to fetch items from
        :return: Iterator over items for this group
        """
        raise NotImplementedError()

    def getItemMap(
        self, group: ItemStorageGroupEnum
    ) -> dict[str, StorageTuple]:
        """Get Item Map

        Get a dictionary with key=key and value=item

        :param group: Group to fetch items from
        :return: A dictionary with all the objects
        True
        """
        raise NotImplementedError()

    @abstractmethod
    def addItem(self, item: StorageTuple):
        """Store Item

        Store an item to the underlying storage. Item needs to be bound to
        the ObjectStorageContext first.

        :param item: Item to store
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def mergeItem(self, item: StorageTuple) -> StorageTuple:
        """Merge Item

        Merges updated item which has the same key as item. Meant to
        be used when an item data is updated (except the key). Invalidates
        instances of the old item with the same key. Returns the updated
        item. Raises `ItemNotFoundError` if item is not stored.


        :param item: Item to be merged
        :return: Bound and stored item
        """
        raise NotImplementedError()

    @abstractmethod
    def updateItemKey(
        self,
        item: StorageTuple,
        newName: str,
    ) -> None:
        """Update Key

        Update the key of a renamed item.

        :param item: Item whose key is to be updated
        :param newName: New name
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def allBranches(self) -> List[str]:
        """All Branches

        Get a list of all user-accessible (Attune) branches

        :return: List of strings
        """
        raise NotImplementedError()

    @abstractmethod
    def allRemotes(self) -> List[tuple[str, str]]:
        """All Remotes

        Get a list of all remotes (name, url pair) defined for the project.

        :return: List of tuples of strings
        """
        raise NotImplementedError()

    @abstractmethod
    def addRemote(self, remote: str, url: str) -> None:
        """Add Remote

        Add a new remote to the project

        :param remote: Name of the remote. Will override existing remote if
        one exists with the same name
        :param url: URL of the remote
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def deleteItem(self, item: StorageTuple):
        """Delete Item

        Delete item from the underlying storage and unbind it

        :param item: Item to be deleted
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def commit(self, msg: str):
        """Create Commit

        Create a raw commit on the working branch

        :param msg: Commit message for the commit
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def checkoutBranch(self, branchName: str) -> None:
        """Checkout Branch

        Checkout (create or switch) to a new (or existing) branch in Git

        :param branchName: Name of the branch to create (or switch if already
        exists)
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def setCommitterSignature(self, name: str, email: str) -> None:
        """Set Committer Signature

        Set the name and email address of the user to use for Squash and Merge
        commits

        :param name: Name of the committer
        :param email: Email of the committer
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def mergeBranches(self, sourceBranch: str) -> None:
        """Merge Branches

        Merge sourceBranch with the currently checked out branch. Reverts and
        does not proceed if led to a merge conflict

        :param sourceBranch: Name of the branch to merge
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def squashAndMergeWorking(self, mergeCommitMessage: str) -> None:
        """Squash and Merge Working

        Creates a commit on the checked branch with all the work since the
        previous commit

        :param mergeCommitMessage: Message to use for the commit
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def pushToRemote(
        self, remote: str, username: str, password: str = None
    ) -> None:
        """Push to Remote

        Push all commits to the URL of the selected remote

        :param remote: Name of the remote to push to
        :param username: Username for remote which has access to the project
        repository
        :param password: Password for the username account
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def getCommits(self) -> Generator[GitCommitTuple, None, None]:
        """Get Commits

        Get all commits for the current checked out branch in reverse
        chronological order

        :return: Generator iterator over commits
        """
        raise NotImplementedError()

    def _cascadeUpdateKey(
        self, itemForKeyUpdate: StorageTuple, newKey: str
    ) -> None:
        """Cascade Update

        This method updates all stored items that point to the key of item
        being updated.

        :param ClassForKeyUpdate: The class of the item being updated
        :param newKey: the new key to use
        :param oldKey: the old key we replacing

        """
        oldKey = itemForKeyUpdate.key

        # Delete references that point to item
        # Handle the simple cases first, that's 1-1
        def handleOneToOne(originRelation, otherItem):
            refFieldName = originRelation.referenceKeyFieldName
            if getattr(otherItem, refFieldName) != oldKey:
                return

            setattr(otherItem, refFieldName, newKey)
            self.mergeItem(otherItem)

        # Handle the one to many with a list of strings case
        def handleOneToManyWithStringList(originRelation, otherItem):
            # This will be a computationally costly exercise.
            refFieldName = originRelation.referenceKeyFieldName
            listVal = getattr(otherItem, refFieldName)
            if not listVal or not oldKey in listVal:
                return

            listVal = [newKey if key == oldKey else key for key in listVal]
            setattr(otherItem, refFieldName, listVal)
            self.mergeItem(otherItem)

        # Handle the one to many with a list of tuples that point to the key
        def handleOneToManyWithTupleList(originRelation, otherItem):
            # This will be a computationally costly exercise.
            refFieldName = originRelation.referenceKeyFieldName
            memberRefFieldName = originRelation.memberReferenceKeyFieldName
            listVal = getattr(otherItem, refFieldName)
            if not listVal:
                return

            updated = False
            for memberTuple in listVal:
                if getattr(memberTuple, memberRefFieldName) == oldKey:
                    setattr(memberTuple, memberRefFieldName, newKey)
                    updated = True

            if updated:
                self.mergeItem(otherItem)

        # Load the relations for this method
        relationsPointingToOurKey: list[
            RelationField
        ] = ObjectStorageContext._relationGraph.getRelationsToStorageTupleKey(
            itemForKeyUpdate.__class__
        )

        # Iterate through each relation
        for originRelation in relationsPointingToOurKey:
            if not originRelation.cascadeOnUpdate:
                continue

            otherItemsIter = filter(
                lambda r: isinstance(r, originRelation.OriginClass),
                self.getItems(originRelation.OriginClass.storageGroup),
            )
            for otherItem in otherItemsIter:
                # If we've deleted it, then move on.
                if otherItem.storageIsExpired:
                    continue

                # handle the four relation cases
                if originRelation.isList:
                    if originRelation.memberReferenceKeyFieldName:
                        handleOneToManyWithTupleList(originRelation, otherItem)
                    else:
                        handleOneToManyWithStringList(originRelation, otherItem)

                else:
                    if originRelation.memberReferenceKeyFieldName:
                        raise NotImplementedError()
                    else:
                        handleOneToOne(originRelation, otherItem)

    def _cascadeDelete(self, itemToDelete: StorageTuple) -> None:
        """Cascade Delete

        This method deletes all stored items that point to the key of item
        being deleted.

        :param: itemToDelete The item being deleted

        """

        # Delete references that point to item
        # Handle the simple cases first, that's 1-1
        def handleOneToOne(originRelation, otherItem):
            refFieldName = originRelation.referenceKeyFieldName
            if getattr(otherItem, refFieldName) != itemToDelete.key:
                return

            if originRelation.cascadeOnDelete:
                self.deleteItem(otherItem)
            else:
                setattr(otherItem, refFieldName, None)
                self.mergeItem(otherItem)

        # Handle the one to many with a list of strings case
        def handleOneToManyWithStringList(originRelation, otherItem):
            # We could just set the entry to None, but I doubt that's useful
            assert originRelation.cascadeOnDelete, "We're only cascade deleting"
            refFieldName = originRelation.referenceKeyFieldName
            listVal = getattr(otherItem, refFieldName)
            if not listVal:
                return

            listVal = [key for key in listVal if key != itemToDelete.key]
            setattr(otherItem, refFieldName, listVal)

        # Handle the one to many with a list of tuples that point to the key
        def handleOneToManyWithTupleList(originRelation, otherItem):
            assert originRelation.cascadeOnDelete, "We're only cascade deleting"
            refFieldName = originRelation.referenceKeyFieldName
            memberRefFieldName = originRelation.memberReferenceKeyFieldName
            listVal = getattr(otherItem, refFieldName)
            if not listVal:
                return

            listVal = [
                memberTuple
                for memberTuple in listVal
                if getattr(memberTuple, memberRefFieldName) != itemToDelete.key
            ]
            setattr(otherItem, refFieldName, listVal)

        # Load the relations for this method
        ClassForDelete = itemToDelete.__class__
        relationsPointingToOurKey: list[
            RelationField
        ] = ObjectStorageContext._relationGraph.getRelationsToStorageTupleKey(
            ClassForDelete
        )

        # Iterate through each relation
        for originRelation in relationsPointingToOurKey:
            otherItemsIter = filter(
                lambda r: isinstance(r, originRelation.OriginClass),
                self.getItems(originRelation.OriginClass.storageGroup),
            )
            for otherItem in otherItemsIter:
                # If we've deleted it, then move on.
                if otherItem.storageIsExpired:
                    continue

                # handle the four relation cases
                if originRelation.isList:
                    if originRelation.memberReferenceKeyFieldName:
                        handleOneToManyWithTupleList(originRelation, otherItem)
                    else:
                        handleOneToManyWithStringList(originRelation, otherItem)

                else:
                    if originRelation.memberReferenceKeyFieldName:
                        raise NotImplementedError()
                    else:
                        handleOneToOne(originRelation, otherItem)



"""Move SQL to script.txt file in Oracle Steps

Revision ID: 58b1656aac40
Revises: 16e34a42a9a1
Create Date: 2022-03-16 15:27:49.806832

"""
import json
from pathlib import Path

from attune_project_api import ObjectStorageContext
from attune_project_api._private.utils import makeStorageKey
from attune_project_api.items.step_tuples.step_sql_oracle_tuple import (
    StepSqlOracleTuple,
)
from attune_project_api.items.step_tuples.step_tuple import StepTuple

# revision identifiers, used by Alembic.
revision = "58b1656aac40"
down_revision = "16e34a42a9a1"
branch_labels = None
depends_on = None


def upgrade(storageContext: ObjectStorageContext):
    try:
        stepsTree = storageContext._getTree(Path("steps"))
    except FileNotFoundError:
        return

    BaseClass = StepTuple

    for itemTree in stepsTree:
        itemPath = Path(
            f"{BaseClass.storageGroup.value}/{makeStorageKey(itemTree.name)}"
        )

        try:
            data = storageContext._readFile((itemPath / "metadata.json"))
            metadataStr = data.decode()
            metadata = json.loads(metadataStr)
        except FileNotFoundError:
            continue

        if metadata["type"] != StepSqlOracleTuple.tupleType():
            continue

        sql = metadata.pop("sql", "")

        # Rewrite the metadata.json
        storageContext._writeFile(
            (itemPath / "metadata.json").as_posix(),
            json.dumps(
                metadata,
                indent=4,
                sort_keys=True,
                separators=(", ", ": "),
            ).encode(),
        )

        # Rewrite the script.txt
        storageContext._writeFile(
            (itemPath / "script.txt").as_posix(), sql.encode()
        )

    storageContext.commit("Moved SQL to script.txt")


def downgrade(storageContext: ObjectStorageContext):
    pass



from abc import ABCMeta
from enum import Enum
from typing import Optional

import markdown
from vortex.Tuple import PolymorphicTupleTypeFieldArg
from vortex.Tuple import TupleField

from attune_project_api import ObjectStorageContext
from attune_project_api import StorageTuple
from attune_project_api.StorageTuple import ItemStorageGroupEnum
from attune_project_api.items import NotZeroLenStr


class StepTupleTypeEnum(Enum):
    GROUP = "com.servertribe.attune.tuples.StepGroupTuple"
    SQL_ORACLE = "com.servertribe.attune.tuples.StepSqlOracleTuple"
    SSH = "com.servertribe.attune.tuples.StepSshTuple"
    SSH_PROMPTED = "com.servertribe.attune.tuples.StepSshPromptedTuple"
    WINRM = "com.servertribe.attune.tuples.StepWinRmTuple"

    PUSH_DESIGN_FILE = "com.servertribe.attune.tuples.StepPushDesignFileTuple"
    PUSH_DESIGN_FILE_COMPILED = (
        "com.servertribe.attune.tuples.StepPushDesignFileCompiledTuple"
    )

    BOOTSTRAP_LINUX = "com.servertribe.attune.tuples.StepBootstrapLinuxTuple"

    TCP_PING = "com.servertribe.attune.tuples.StepTcpPingTuple"

    PROJECT_LINK = "com.servertribe.attune.tuples.StepProjectLinkTuple"


@ObjectStorageContext.registerItemClass
class StepTuple(StorageTuple, metaclass=ABCMeta):
    __tupleArgs__ = (PolymorphicTupleTypeFieldArg("type"),)
    __group__ = ItemStorageGroupEnum.Step

    name: NotZeroLenStr = TupleField()
    comment: Optional[str] = TupleField(defaultValue="")
    enabled: bool = TupleField(defaultValue=True)
    type: NotZeroLenStr = TupleField()

    def parameters(self) -> list["ParameterTuple"]:
        """Parameters

        return a list of parameters linked to from this step

        """
        # This is an abstract kind of method, but we can't have an abstract
        # class because StepTuple() is constructed to send to the UI
        raise NotImplementedError()

    def scriptReferences(self) -> list[str]:
        """Script References

        return a list of script references within this step.

        """
        # This is an abstract kind of method, but we can't have an abstract
        # class because StepTuple() is constructed to send to the UI
        raise NotImplementedError()

    def makeCommentHtml(self, topHeaderNum: int = 1):
        html = markdown.markdown(
            self.comment,
            extensions=[
                "markdown.extensions.tables",
                "markdown.extensions.fenced_code",
                "markdown.extensions.sane_lists",
            ],
        )

        highestHeader = 6
        for h in reversed(range(1, 10)):
            if "<h%s>" % h in html:
                highestHeader = h

        # If highest header is <
        headerDelta = topHeaderNum - highestHeader
        if headerDelta == 0:
            return html

        for h in reversed(range(1, 10)):
            html = html.replace("<h%s>" % h, "<h%s>" % (h + headerDelta))
            html = html.replace("</h%s>" % h, "</h%s>" % (h + headerDelta))

        return html



import logging
import os
import sqlite3

from alembic.config import Config
from alembic.runtime.migration import MigrationContext
from alembic.script import ScriptDirectory
from sqlalchemy import create_engine

from attune_project_api import ObjectStorageContext


logger = logging.getLogger(__name__)


def runMigrationsForStorageContext(storageContext: ObjectStorageContext):
    currentRev = storageContext.getRevision()

    # Alembic requires a DB to read the version from. We mock an in-memory DB
    # that has a table `alembic_version` with a single value as the current rev
    db = sqlite3.connect("file::memory:")
    cur = db.cursor()
    cur.execute("CREATE TABLE alembic_version(version_num varchar)")
    # What if someone sets currentRev to "Robert'); DROP TABLE users;--" ?
    cur.execute(
        f"INSERT INTO alembic_version (version_num) VALUES ('{currentRev}')"
    )
    db.commit()

    # Since sqlite3 can have only one memory DB per thread, this `:memory:` DB
    # will be the same as the one created previously
    engine = create_engine("sqlite:///:memory:")

    # Load the migration scripts and
    codeDir = os.path.dirname(os.path.realpath(__file__))
    configFilePath = os.path.join(codeDir, "alembic.ini")
    migrationsDir = os.path.join(codeDir, "alembic_migrations")

    config = Config(file_=configFilePath)
    config.set_main_option("script_location", migrationsDir)
    script = ScriptDirectory.from_config(config)
    latestRev = script.get_heads()[0]

    # The migration context expects a function to return an iterator of
    # revisions from a current revision (`revision`) to the destination
    # revision `latestRev`. The `script._upgrade_revs` method provides such
    # an iterator built from the scripts under `versions`
    def migrations_fn(revision, _ctx):
        return script._upgrade_revs(latestRev, revision)

    with engine.connect() as conn:
        context = MigrationContext.configure(
            connection=conn,
            opts={"transactional_ddl": False, "fn": migrations_fn},
        )

        if currentRev != latestRev:
            logger.info("Running migrations for Project API")
            context.run_migrations(storageContext=storageContext)
            storageContext.setRevision(latestRev)
            storageContext.squashAndMergeWorking(
                f"Migrate to {latestRev} revision"
            )


def getLatestRevision() -> str:
    codeDir = os.path.dirname(os.path.realpath(__file__))
    configFilePath = os.path.join(codeDir, "alembic.ini")
    migrationsDir = os.path.join(codeDir, "alembic_migrations")

    config = Config(file_=configFilePath)
    config.set_main_option("script_location", migrationsDir)
    script = ScriptDirectory.from_config(config)
    return script.get_heads()[0]



from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from attune_project_api import ObjectStorageContext
from attune_project_api import ParameterTuple
from attune_project_api.RelationField import RelationField
from attune_project_api.items import NotZeroLenStr
from attune_project_api.items.step_tuples import addStepDeclarative
from attune_project_api.items.step_tuples.step_tuple import StepTuple
from attune_project_api.items.step_tuples.step_tuple import StepTupleTypeEnum


@ObjectStorageContext.registerItemClass
@addStepDeclarative("Setup Linux SSH Keys")
@addTupleType
class StepBootstrapLinuxTuple(StepTuple):
    __tupleType__ = StepTupleTypeEnum.BOOTSTRAP_LINUX.value

    serverKey: NotZeroLenStr = TupleField()
    osCredKey: NotZeroLenStr = TupleField()

    server: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="serverKey",
    )
    osCred: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="osCredKey",
    )

    def parameters(self) -> list["ParameterTuple"]:
        return [self.server, self.osCred]

    def scriptReferences(self) -> list[str]:
        return []



import re

stepNiceNames = {}
STEP_TYPES = {}


def addStepDeclarative(niceName):
    def f(cls):
        cls.NICE_NAME = niceName
        stepNiceNames[cls.tupleType()] = niceName
        STEP_TYPES[cls.tupleType()] = cls
        return cls

    return f


def extractTextPlaceholders(text):
    # NOTE, Duplicated client side, AttunePlaceholderMod.js,
    # $scope.textPhName = function () { ...
    return re.findall(r"(?<![$]){([A-Za-z0-9_.]+)}", text)


def loadStorageStepTuples():
    from . import step_sql_oracle_tuple
    from . import step_tuple
    from . import step_bootstrap_linux_tuple
    from . import step_push_design_file_compiled_tuple
    from . import step_group_tuple
    from . import step_push_design_file_tuple
    from . import step_ssh_tuple
    from . import step_ssh_prompted_tuple
    from . import step_winrm_tuple
    from . import step_tcp_ping_tuple
    from . import step_project_link_tuple



from attune_project_api import StorageTuple
from attune_project_api.Exceptions import InvalidReferenceError
from attune_project_api.Exceptions import NonUniqueScriptRefError
from attune_project_api.ObjectStorageContext import ObjectStorageContext
from attune_project_api.ObjectStorageContext import TupleValidatorABC
from attune_project_api.items.parameter_tuple import ParameterTuple
from attune_project_api.items.step_tuples.step_tuple import StepTuple


@ObjectStorageContext.registerValidator
class ParameterScriptRefValidator(TupleValidatorABC):
    """ParameterTuple Name Validator

    ParameterTuple names are checked for uniqueness by UniqueNameValidator.
    They should also result in unique script references

    """

    def validate(
        self,
        context: "ObjectStorageContext",
        item: ParameterTuple,
    ) -> None:
        if not isinstance(item, ParameterTuple):
            return

        scriptRef = item.textName
        paramsWithScriptRef = filter(
            lambda p: p.textName == scriptRef and p.name != item.name,
            context.getItems(ParameterTuple.storageGroup),
        )

        if len(list(paramsWithScriptRef)) != 0:
            raise NonUniqueScriptRefError(
                f"ParameterTuple with script reference {scriptRef} "
                f"already exists"
            )


@ObjectStorageContext.registerValidator
class StepDeployConfigParamValidator(TupleValidatorABC):
    def validate(
        self, context: ObjectStorageContext, item: StorageTuple
    ) -> None:
        from attune_project_api.items.step_tuples.step_push_design_file_compiled_tuple import (
            StepPushDesignFileCompiledParamTuple,
            StepPushDesignFileCompiledTuple,
        )

        if item is None or not isinstance(
            item, StepPushDesignFileCompiledParamTuple
        ):
            return

        placeholder = context.getItem(
            ParameterTuple.storageGroup, item.placeholderName
        )
        configStep = context.getItem(StepTuple.storageGroup, item.configstepKey)

        if (
            configStep is None
            or not isinstance(configStep, StepPushDesignFileCompiledTuple)
        ) and placeholder is None:
            raise InvalidReferenceError(
                f"A ParameterTuple with {item.placeholderName} and StepTuple"
                f" with {item.configstepKey} does not exist"
            )
        elif configStep is None or not isinstance(
            configStep, StepPushDesignFileCompiledTuple
        ):
            raise InvalidReferenceError(
                f"A StepTuple with {item.configstepKey} does not exist"
            )
        elif placeholder is None:
            raise InvalidReferenceError(
                f"A ParameterTuple with {item.placeholderName} does not exist"
            )



from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType


@addTupleType
class GitCommitTuple(Tuple):
    __tupleType__ = "attune.auto.GitCommitTuple"

    timestamp: str = TupleField()
    hash: str = TupleField()
    authorName: str = TupleField()
    authorEmail: str = TupleField()
    message: str = TupleField()



import hashlib
import logging
from datetime import datetime
from pathlib import Path

import pytz

logger = logging.getLogger(__name__)


def makeArchiveMd5(path: Path) -> str:
    startTime = datetime.now(pytz.utc)
    BUF_SIZE = 5 * 1024 * 1024
    md5 = hashlib.md5()

    with open(path, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)

    md5 = md5.hexdigest()

    logger.info(
        "We hashed file %s in %s",
        path,
        datetime.now(pytz.utc) - startTime,
    )

    return md5



from typing import Optional

from attune_project_api import StorageTuple
from attune_project_api.StorageTuple import StorageMemberTuple


class RelationField:
    def __init__(
        self,
        ForeignClass: type[StorageTuple],
        referenceKeyFieldName: str,
        cascadeOnDelete: bool = False,
        isList: bool = False,
        memberReferenceKeyFieldName: Optional[str] = None,
        cascadeOnUpdate=True,
    ):
        """Relation

        :param: memberReferenceKeyFieldName
         The dot reference field name creates a relation out of child objects
         for example

         class OtherTuple(StorageTuple):
             key = TupleField()

         class DotTuple(StorageMemberTuple):
             otherKey = TupleField()

             other = RelationField(
                tupleType=OtherTuple.tupleType(),
                referenceKeyFieldName="otherKey",
                cascadeOnDelete=True,
            )

         class Thing2Tuple(StorageTuple):
             key = TupleField()

             dotLinks:list[DotTuple] = TupleField()

             otherThings = Relation(
                tupleType=DotTuple.tupleType(),
                referenceKeyFieldName="dotLinks",
                isList=True,
                memberReferenceKeyFieldName="otherKey
            )



        """
        self.ForeignClass = ForeignClass
        self.referenceKeyFieldName = referenceKeyFieldName
        self.cascadeOnDelete = cascadeOnDelete
        self.isList = isList
        self.memberReferenceKeyFieldName = memberReferenceKeyFieldName
        self.cascadeOnUpdate = cascadeOnUpdate
        self.OriginClass = None

        if self.isList and not self.cascadeOnDelete:
            raise Exception("Relations expect cascade on delete for lists")

    def _setOriginClass(self, OriginClass: type[StorageTuple]) -> None:
        self.OriginClass = OriginClass

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None):
        if instance is None:
            return

        isMemberTuple = isinstance(instance, StorageMemberTuple)
        if isMemberTuple:
            context = instance.storageParent.storageContext
        else:
            context = instance.storageContext

        if not context:
            raise ValueError(f"ObjectStorageContext not bound")

        if getattr(instance, self.referenceKeyFieldName) is None:
            return None

        # It doesn't matter if we're a list or not, the member tuple level
        # must be a 1 to 1 relationship.
        if isMemberTuple:
            # From the member tuples perspective,
            # use referenceKeyFieldName, not memberReferenceKeyFieldName
            return context.getItem(
                self.ForeignClass.storageGroup,
                getattr(instance, self.referenceKeyFieldName),
            )

        # instance is the main StorageTuple, and this is not a list
        # handle this simple case and return
        if not self.isList:
            if self.memberReferenceKeyFieldName:
                raise NotImplementedError(
                    "Member Tuples that are not a list are not yet implemented"
                )
            return context.getItem(
                self.ForeignClass.storageGroup,
                getattr(instance, self.referenceKeyFieldName),
            )

        # We have a list, get it from the instance
        values = getattr(instance, self.referenceKeyFieldName)
        if not values:
            return []

        # If we have this set, then we have a list of tuples, each with a
        # variable that points to a key, convert that to a simple list
        if self.memberReferenceKeyFieldName:
            # Create a set of keys from our member tuples variable
            values = [
                getattr(memberTuple, self.memberReferenceKeyFieldName)
                for memberTuple in values
            ]

        # Using the set above, return a list that matches our list,
        # Doing it this way will maintain the list length, order and any
        # duplicates, and raise any exceptions for missing keys
        return [
            context.getItem(self.ForeignClass.storageGroup, key)
            for key in values
        ]



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""
import re
from typing import Any
from typing import Optional

from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from . import NotZeroLenStr
from .. import ObjectStorageContext
from ..StorageTuple import ItemStorageGroupEnum
from ..StorageTuple import StorageTuple


def convertParameterNameToScriptRef(name):
    # NOTE, Duplicated client side, AttunePlaceholderMod.js,
    # $scope.textPhName = function () { ...
    return re.sub(r"[^A-Za-z0-9_]", "", name).lower()


@ObjectStorageContext.registerItemClass
@addTupleType
class ParameterTuple(StorageTuple):
    __tupleType__ = "c.s.s.b.Placeholder"
    __group__ = ItemStorageGroupEnum.Parameter

    # GENERIC_SERVER = ValueTypeEnum.GENERIC_SERVER.value
    # LIN_SERVER = ValueTypeEnum.LIN_SERVER.value
    # WIN_SERVER = ValueTypeEnum.WIN_SERVER.value
    # SERVER_LIST = ValueTypeEnum.SERVER_LIST.value
    # GENERIC_CRED = ValueTypeEnum.GENERIC_CRED.value
    # WIN_OS_CRED = ValueTypeEnum.WIN_OS_CRED.value
    # LIN_OS_CRED = ValueTypeEnum.LIN_OS_CRED.value
    # SQL_CRED = ValueTypeEnum.SQL_CRED.value
    # IP4_SUBNET = ValueTypeEnum.IP4_SUBNET.value
    # TEXT = ValueTypeEnum.TEXT.value

    # Temporary to get it working with Attune
    GENERIC_SERVER = "c.s.s.b.phv.Server"
    LIN_SERVER = "c.s.s.b.phv.LinuxServer"
    WIN_SERVER = "c.s.s.b.phv.WindowsServer"
    SERVER_LIST = "c.s.s.b.phv.ServerList"
    GENERIC_CRED = "c.s.s.b.phv.OsCred"
    WIN_OS_CRED = "c.s.s.b.phv.WinOsCred"
    LIN_OS_CRED = "c.s.s.b.phv.LinOsCred"
    SQL_CRED = "c.s.s.b.phv.SqlCred"
    IP4_SUBNET = "c.s.s.b.phv.Ip4Subnet"
    TEXT = "c.s.s.b.phv.Text"

    uiData: Optional[dict] = TupleField({}, jsonExclude=True)

    key: NotZeroLenStr = TupleField()
    type: NotZeroLenStr = TupleField(defaultValue=LIN_SERVER)
    name: NotZeroLenStr = TupleField()
    showInPlan: bool = TupleField(defaultValue=True)
    comment: Optional[str] = TupleField()

    def __str__(self):
        return self.name

    __repr__ = __str__

    @property
    def textName(self):
        return convertParameterNameToScriptRef(self.name)

    @property
    def variableValueMapKey(self):
        return self.name, self.type


# This must match attune-ui/src/app/parameter/input-constants.ts
ParameterTypeNames = {
    ParameterTuple.SERVER_LIST: "Node List",
    ParameterTuple.TEXT: "Text",
    ParameterTuple.GENERIC_SERVER: "Basic Node",
    ParameterTuple.GENERIC_CRED: "Basic Credential",
    ParameterTuple.WIN_SERVER: "Windows Node",
    ParameterTuple.WIN_OS_CRED: "Windows Credential",
    ParameterTuple.LIN_SERVER: "Linux/Unix Node",
    ParameterTuple.LIN_OS_CRED: "Linux/Unix Credential",
    ParameterTuple.IP4_SUBNET: "Network IPv4 Subnet",
    ParameterTuple.SQL_CRED: "Oracle SQL Credential",
}
niceParameterNames = ParameterTypeNames
niceValueNames = ParameterTypeNames



from collections import namedtuple

from attune_project_api.RelationField import RelationField

BackReference = namedtuple(
    "BackReference",
    [
        "origin",
        "referenceKeyFieldName",
        "cascade",
        "isList",
        "memberReferenceKeyFieldName",
    ],
)

Cascade = namedtuple(
    "Cascade", ["attribName", "cascadedOn", "memberReferenceKeyFieldName"]
)


class RelationReferences:
    def __init__(self):
        self.back_references: list[BackReference] = []
        self.cascades: list[Cascade] = []


class RelationGraph:
    def __init__(self):
        self.__stagedRelations: list[RelationField] = []

    def getRelationsToStorageTupleKey(self, Class_) -> list[RelationField]:
        return list(
            set(
                filter(
                    lambda r: issubclass(Class_, r.ForeignClass),
                    self.__stagedRelations,
                )
            )
        )

    def add(self, relation: RelationField):
        if self.__stagedRelations is None:
            raise Exception("The RelationGraph has already been compiled")
        self.__stagedRelations.append(relation)



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""

from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from . import addStepDeclarative
from .step_tuple import StepTupleTypeEnum
from .. import NotZeroLenStr
from ... import ParameterTuple
from ... import StepTuple
from ...ObjectStorageContext import ObjectStorageContext
from ...RelationField import RelationField


@ObjectStorageContext.registerItemClass
@addStepDeclarative("Tcp Ping")
@addTupleType
class StepTcpPingTuple(StepTuple):
    __tupleType__ = StepTupleTypeEnum.TCP_PING.value

    serverKey: NotZeroLenStr = TupleField()
    tcpPort: int = TupleField()
    preWait: int = TupleField(defaultValue=0)
    minDowntime: int = TupleField(defaultValue=0)
    postWait: int = TupleField(defaultValue=0)

    server: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="serverKey",
    )

    def parameters(self) -> list["ParameterTuple"]:
        return [self.server]

    def scriptReferences(self) -> list[str]:
        return []



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""

from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from . import addStepDeclarative
from .step_tuple import StepTupleTypeEnum
from ... import ParameterTuple
from ... import StepTuple
from ...ObjectStorageContext import ObjectStorageContext


@ObjectStorageContext.registerItemClass
@addStepDeclarative("Project Link")
@addTupleType
class StepProjectLinkTuple(StepTuple):
    __tupleType__ = StepTupleTypeEnum.PROJECT_LINK.value

    projectKey: str = TupleField()
    blueprintKey: str = TupleField()
    pullUrl: str = TupleField()

    def parameters(self) -> list["ParameterTuple"]:
        return []

    def scriptReferences(self) -> list[str]:
        return []



from pathlib import Path

from attune_project_api.ObjectStorageContext import ArchiveFileInfo
from attune_project_api._private.archive_format import guessArchiveFormat
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_7Z,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR_BZ2,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR_GZ,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR_XZ,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_ZIP,
)

__ARCHIVE_LISTER_BY_FORMAT_KEY = {}


def __registerArchiveLister(key: str):
    def wrapper(listerMethod):
        __ARCHIVE_LISTER_BY_FORMAT_KEY[key] = listerMethod
        return listerMethod

    return wrapper


def canListFiles(archiveFormatKey: str) -> bool:
    return archiveFormatKey in __ARCHIVE_LISTER_BY_FORMAT_KEY


def listArchiveFiles(path: Path) -> list[ArchiveFileInfo]:
    format = guessArchiveFormat(path)
    lister = __ARCHIVE_LISTER_BY_FORMAT_KEY[format.key]
    return lister(path)


@__registerArchiveLister(ARCHIVE_FORMAT_KEY_7Z)
def _list7z(path: Path) -> list[ArchiveFileInfo]:
    from py7zr import SevenZipFile

    with SevenZipFile(path, "r") as reader:
        files = reader.files
        return [
            ArchiveFileInfo(Path(m.filename), m.uncompressed)
            for m in files
            if not m.is_directory
        ]


@__registerArchiveLister(ARCHIVE_FORMAT_KEY_ZIP)
def _listZip(path: Path) -> list[ArchiveFileInfo]:
    from zipfile import ZipFile, ZIP_DEFLATED, ZipInfo

    with ZipFile(path, "r", ZIP_DEFLATED) as reader:
        infos: list[ZipInfo] = reader.infolist()
        return [
            ArchiveFileInfo(Path(m.filename), m.file_size)
            for m in infos
            if not m.is_dir()
        ]


@__registerArchiveLister(ARCHIVE_FORMAT_KEY_TAR)
@__registerArchiveLister(ARCHIVE_FORMAT_KEY_TAR_GZ)
@__registerArchiveLister(ARCHIVE_FORMAT_KEY_TAR_BZ2)
@__registerArchiveLister(ARCHIVE_FORMAT_KEY_TAR_XZ)
def _listTarFile(path: Path) -> list[ArchiveFileInfo]:
    from tarfile import TarFile, TarInfo, DIRTYPE

    with TarFile.open(path, "r") as reader:
        members: list[TarInfo] = reader.getmembers()
        return [
            ArchiveFileInfo(Path(m.path), m.size)
            for m in members
            if m.type != DIRTYPE
        ]



"""Attune v5 Starting Point

Revision ID: 16e34a42a9a1
Revises: 
Create Date: 2022-03-16 15:26:28.813729

"""

# revision identifiers, used by Alembic.
from attune_project_api import ObjectStorageContext

revision = "16e34a42a9a1"
down_revision = None
branch_labels = None
depends_on = None


def upgrade(storageContext: ObjectStorageContext):
    pass


def downgrade(storageContext: ObjectStorageContext):
    pass



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""
import logging
from pathlib import Path
from typing import Optional

from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from attune_project_api.ObjectStorageContext import ArchiveFileContent
from attune_project_api.ObjectStorageContext import ArchiveFileInfo
from attune_project_api._contexts.GitObjectStorageContext import (
    GitObjectStorageContext,
)
from attune_project_api._private.archive_format import ArchiveFormat
from attune_project_api._private.archive_format import (
    InvalidArchiveFormatException,
)
from attune_project_api._private.archive_format import archiveFormatByKey
from attune_project_api._private.archive_format import guessArchiveFormat
from attune_project_api._private.archive_hasher import makeArchiveMd5
from attune_project_api.items.file_archive_tuples.file_archive_tuple import (
    FileArchiveTuple,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_UNKNOWN,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    FileArchiveTupleTypeEnum,
)

logger = logging.getLogger(__name__)


@GitObjectStorageContext.registerItemClass
@addTupleType
class FileArchiveLargeTuple(FileArchiveTuple):
    """Archive Large Tuple

    The contents of these archives are not stored in Attune.


    """

    __tupleType__ = FileArchiveTupleTypeEnum.LARGE.value
    __ARCHIVE_FILE_NAME = "large_file_archive"

    #: This is the md5 hash of the last downloaded file
    md5: Optional[str] = TupleField()

    #: This is the size of the last downloaded file
    size: Optional[int] = TupleField()

    #: This is the name of the file that was last manually uploaded
    fileName: Optional[str] = TupleField()

    #: This is the key of the format of the archive
    # See archive_format.py, ARCHIVE_FORMAT_KEY_*
    formatKey: Optional[str] = TupleField()

    #: The uri to pull this file from.
    # This can be HTTP(S) or SSH
    remoteUri: Optional[str] = TupleField()

    @classmethod
    @property
    def niceName(cls) -> str:
        # This is meant to be an abstract method
        return "Large Archives"

    @property
    def archiveFormat(self) -> Optional[ArchiveFormat]:
        return archiveFormatByKey.get(self.formatKey, None)

    @property
    def archivePath(self) -> Path:
        return (
            self.storageContext.getItemLargeFilesPath(
                self.storageGroup, self.key
            )
            / self.__ARCHIVE_FILE_NAME
        )

    @property
    def listFiles(self) -> list[ArchiveFileInfo]:
        path = self.archivePath
        if not path.exists():
            return []
        from attune_project_api._private.archive_list_files import (
            listArchiveFiles,
        )

        return listArchiveFiles(self.archivePath)

    def getFileContent(self, path: Path) -> ArchiveFileContent:
        archivePath = self.archivePath
        if not archivePath.exists():
            raise FileNotFoundError(
                "Archive file %s doesn't exist" % archivePath
            )

        from attune_project_api._private.archive_get_file_content import (
            getArchiveFileContent,
        )

        return getArchiveFileContent(archivePath, path)

    @property
    def isListable(self) -> bool:
        from attune_project_api._private.archive_list_files import canListFiles

        return canListFiles(self.formatKey)

    @property
    def isTarFile(self) -> bool:
        """Is Tar File"""
        from attune_project_api._private.archive_format import isTarFile

        return isTarFile(self.formatKey)

    @property
    def is7zFile(self) -> bool:
        """Is 7z File"""
        from attune_project_api._private.archive_format import is7zFile

        return is7zFile(self.formatKey)

    def makeFileName(self) -> str:
        return self._makeFileName(self.formatKey)

    def refreshArchiveInfo(self) -> None:
        filePath = self.archivePath

        if not filePath.exists():
            self.md5 = None
            self.size = None
            self.formatKey = None
            return

        self.size = filePath.stat().st_size
        self.md5 = makeArchiveMd5(filePath)
        try:
            self.formatKey = guessArchiveFormat(filePath).key
        except InvalidArchiveFormatException:
            self.formatKey = ARCHIVE_FORMAT_KEY_UNKNOWN

        if not self.fileName:
            self.fileName = self.makeFileName()



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""
import logging
from datetime import datetime
from io import BytesIO
from pathlib import Path
from zipfile import ZIP_DEFLATED
from zipfile import ZIP_STORED

import pytz
from pytmpdir.spooled_named_temporary_file import SpooledNamedTemporaryFile
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.DeferUtil import noMainThread
from vortex.Tuple import addTupleType

from attune_project_api.ObjectStorageContext import VersionedFileContent
from attune_project_api.ObjectStorageContext import VersionedFileInfo
from attune_project_api._contexts.GitObjectStorageContext import (
    GitObjectStorageContext,
)
from attune_project_api.items.file_archive_tuples.file_archive_tuple import (
    FileArchiveTuple,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_ZIP,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    FileArchiveTupleTypeEnum,
)

logger = logging.getLogger(__name__)


@GitObjectStorageContext.registerItemClass
@addTupleType
class FileArchiveVersionedTuple(FileArchiveTuple):
    """Archive Versioned Tuple

    This archive tuple stores small version controlled archives.
    They are never archived, they live extracted in the git object repository
    """

    # We need to enforce this because this data is loaded entirely in memory
    MAX_VERSIONED_FILE_SIZE = 1 * 1024 * 1024
    MAX_ARCHIVE_VERSION_REPLACE_SIZE = 10 * 1024 * 1024

    __tupleType__ = FileArchiveTupleTypeEnum.VERSIONED.value

    @classmethod
    @property
    def niceName(cls) -> str:
        # This is meant to be an abstract method
        return "Version Controlled Files"

    def __checkSize(self, data):
        if len(data) > self.MAX_VERSIONED_FILE_SIZE:
            raise Exception(
                "Only files up to 1mb can be version controlled,"
                "use the Large files type instead"
            )

    def writeFile(self, path: Path, data: bytes) -> None:
        self.__checkSize(data)
        self.storageContext.writeItemVersionedFile(
            self.storageGroup, self.key, path, data
        )

    def readFile(self, path: Path) -> bytes:
        data = self.storageContext.readItemVersionedFile(
            self.storageGroup, self.key, path
        )
        self.__checkSize(data)
        return data

    @property
    def listFiles(self) -> list[VersionedFileInfo]:
        return self.storageContext.listItemVersionedFiles(
            self.storageGroup, self.key
        )

    def getFileContent(self, path: Path) -> VersionedFileContent:
        return self.storageContext.getItemVersionedFileContent(
            self.storageGroup, self.key, path
        )

    def setFileContent(self, path: Path, data: bytes, executable: bool) -> None:
        self.storageContext.setItemVersionedFileContent(
            self.storageGroup, self.key, path, data, executable
        )

    def moveFile(self, fromPath: Path, toPath: Path) -> None:
        self.storageContext.moveItemVersionedFile(
            self.storageGroup, self.key, fromPath, toPath
        )

    def moveDirectory(self, fromPath: Path, toPath: Path) -> None:
        self.storageContext.moveItemVersionedDirectory(
            self.storageGroup, self.key, fromPath, toPath
        )

    def deleteFile(self, path: Path) -> None:
        self.storageContext.deleteItemVersionedFile(
            self.storageGroup, self.key, path
        )

    def deleteDirectory(self, path: Path = Path()) -> None:
        self.storageContext.deleteItemVersionedDirectory(
            self.storageGroup, self.key, path
        )

    def makeFileName(self, formatKey: str = ARCHIVE_FORMAT_KEY_TAR) -> str:
        assert formatKey in (
            ARCHIVE_FORMAT_KEY_ZIP,
            ARCHIVE_FORMAT_KEY_TAR,
        ), "formatKey should must be tar or zip"

        return self._makeFileName(formatKey)

    @deferToThreadWrapWithLogger(logger)
    def makeZipAsync(self, deflate: bool = False) -> SpooledNamedTemporaryFile:
        return self.makeZip(deflate=deflate)

    def makeZip(self, deflate: bool = False) -> SpooledNamedTemporaryFile:
        noMainThread()
        now = datetime.now(pytz.utc)

        tmpFile = SpooledNamedTemporaryFile()
        # Create the zip file
        from zipfile import ZipFile, ZipInfo

        with ZipFile(
            tmpFile, "w", compression=(ZIP_DEFLATED if deflate else ZIP_STORED)
        ) as zip:
            for fileInfo in self.listFiles:
                content = self.getFileContent(fileInfo.path)
                zip.writestr(
                    ZipInfo(str(fileInfo.path), date_time=now.timetuple()),
                    content.data,
                )

        tmpFile.seek(0)
        return tmpFile

    @deferToThreadWrapWithLogger(logger)
    def makeTarAsync(self) -> SpooledNamedTemporaryFile:
        return self.makeTar()

    def makeTar(self) -> SpooledNamedTemporaryFile:
        noMainThread()
        now = datetime.now(pytz.utc)

        tmpFile = SpooledNamedTemporaryFile()

        # Create the zip file
        from tarfile import TarFile, TarInfo

        with TarFile.open(
            fileobj=tmpFile, mode="w", bufsize=1024 * 1024
        ) as tar:
            for fileInfo in self.listFiles:
                content = self.getFileContent(fileInfo.path)

                fileObj = BytesIO(content.data)
                fileObj.seek(0)

                tarInfo = TarInfo(str(fileInfo.path))
                tarInfo.size = len(content.data)
                tarInfo.mtime = now.timestamp()

                tar.addfile(tarInfo, fileObj)

        tmpFile.seek(0)
        return tmpFile



from typing import Annotated


# Commonly used type annotations for TupleFields
from attune_project_api.TupleFieldValidators import NotFalsyValidator

NotZeroLenStr = Annotated[str, NotFalsyValidator()]


def loadStorageTuples():
    from . import parameter_tuple

    from .step_tuples import loadStorageStepTuples

    loadStorageStepTuples()

    from attune_project_api.items.file_archive_tuples import (
        loadStorageFileArchiveTuples,
    )

    loadStorageFileArchiveTuples()



import json
import logging
import shutil
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Iterator
from typing import Optional

import pygit2
import pytz
from twisted.internet.defer import DeferredList
from twisted.internet.defer import DeferredSemaphore
from twisted.internet.defer import inlineCallbacks
from vortex.Tuple import Tuple

from attune_project_api.items.step_tuples.step_sql_oracle_tuple import (
    StepSqlOracleTuple,
)
from attune_project_api.migration import runMigrationsForStorageContext
from .git_lib_mixin import GitLibMixin
from ..Exceptions import ItemNotFoundError
from ..Exceptions import ProjectValidationError
from ..ObjectStorageContext import ContextProjectInfo
from ..ObjectStorageContext import ObjectStorageContext
from ..ObjectStorageContext import ObjectStorageContextFileMixin
from ..ObjectStorageContext import VersionedFileContent
from ..ObjectStorageContext import VersionedFileInfo
from ..StorageTuple import ItemStorageGroupEnum
from ..StorageTuple import StorageTuple
from .._private.large_file_http_downloader import LargeFileHttpDownloader
from .._private.utils import makeStorageKey
from ..items.file_archive_tuples.file_tuple_constants import (
    FileArchiveTupleTypeEnum,
)


logger = logging.getLogger(__name__)

DIR_CREATE_MODE = 0o700


class GitObjectStorageContextFileMixin(ObjectStorageContextFileMixin):
    CONTENTS_PATH = "contents"

    # ---------------
    # List the abstract methods we depend on

    # @abstractmethod
    # @property
    # def projectPath(self) -> Path:
    #     pass
    #
    # @abstractmethod
    # def _readFile(self, path: Path) -> bytes:
    #     pass
    #
    # @abstractmethod
    # def _writeFile(self, path: Path, data: bytes) -> None:
    #     pass
    #
    # @abstractmethod
    # def _getTree(self, path: Path) -> pygit2.Tree:
    #     pass

    # ---------------
    # Start our implementation

    @property
    def _unversionedFilesPath(self) -> Path:
        return self.projectPath / "file_storage"

    def __makeVersionedPath(self, group: ItemStorageGroupEnum, key: str):
        return Path(f"{group.value}/{makeStorageKey(key)}") / self.CONTENTS_PATH

    def readItemVersionedFile(
        self, group: ItemStorageGroupEnum, key: str, path: Path
    ) -> bytes:
        return self._readFile(self.__makeVersionedPath(group, key) / path)

    def writeItemVersionedFile(
        self, group: ItemStorageGroupEnum, key: str, path: Path, data: bytes
    ) -> None:
        self._writeFile(self.__makeVersionedPath(group, key) / path, data)

    def listItemVersionedFiles(
        self, group: ItemStorageGroupEnum, key: str
    ) -> list[VersionedFileInfo]:
        infos = []
        try:
            tree = self._getTree(self.__makeVersionedPath(group, key))
        except FileNotFoundError:
            logger.debug(
                "No content files found for Group %s, Key %s", group, key
            )
            return []

        def recurse(tree, parentPath: Path):
            for item in tree:
                if isinstance(item, pygit2.Tree):
                    recurse(item, parentPath / item.name)
                else:
                    assert isinstance(item, pygit2.Object)
                    infos.append(
                        VersionedFileInfo(
                            path=parentPath / item.name,
                            size=item.size,
                            executable=item.filemode
                            & pygit2.GIT_FILEMODE_BLOB_EXECUTABLE,
                            sha1=item.hex,
                        )
                    )

        recurse(tree, Path())
        return infos

    def getItemVersionedFileContent(
        self, group: ItemStorageGroupEnum, key: str, path: Path
    ) -> Optional[VersionedFileContent]:
        try:
            tree = self._getTree(
                self.__makeVersionedPath(group, key) / path.parent
            )
        except FileNotFoundError:
            logger.error(
                "No file found for Group %s, Key %s, Path %s", group, key, path
            )
            raise

        if path.name not in tree:
            logger.error(
                "No file found for Group %s, Key %s, Path %s", group, key, path
            )
            raise FileNotFoundError()

        fileObject = tree / path.name
        if not isinstance(fileObject, pygit2.Object):
            logger.error(
                "Found a directory, not a file for Group %s, Key %s, Path %s",
                group,
                key,
                path,
            )
            raise FileNotFoundError()

        return VersionedFileContent(
            path=path,
            data=fileObject.data,
            executable=fileObject.filemode
            & pygit2.GIT_FILEMODE_BLOB_EXECUTABLE,
            sha1=fileObject.hex,
        )

    def setItemVersionedFileContent(
        self,
        group: ItemStorageGroupEnum,
        key: str,
        path: Path,
        data: bytes,
        executable: bool,
    ) -> None:
        treePath = self.__makeVersionedPath(group, key) / path
        mode = (
            pygit2.GIT_FILEMODE_BLOB_EXECUTABLE
            if executable
            else pygit2.GIT_FILEMODE_BLOB
        )

        self._writeFile(
            treePath.as_posix(),
            data,
            mode=mode,
        )

    def moveItemVersionedFile(
        self,
        group: ItemStorageGroupEnum,
        key: str,
        fromPath: Path,
        toPath: Path,
    ) -> None:
        fromContentPath = self.__makeVersionedPath(group, key) / fromPath
        toContentPath = self.__makeVersionedPath(group, key) / toPath
        self._moveFile(fromContentPath, toContentPath)

    def moveItemVersionedDirectory(
        self,
        group: ItemStorageGroupEnum,
        key: str,
        fromPath: Path,
        toPath: Path,
    ) -> None:
        fromContentPath = self.__makeVersionedPath(group, key) / fromPath
        toContentPath = self.__makeVersionedPath(group, key) / toPath
        self._moveDirectory(fromContentPath, toContentPath)

    def deleteItemVersionedFile(
        self, group: ItemStorageGroupEnum, key: str, path: Path
    ) -> None:
        contentPath = self.__makeVersionedPath(group, key) / path
        self._deleteFile(contentPath)

    def deleteItemVersionedDirectory(
        self, group: ItemStorageGroupEnum, key: str, path: Path
    ) -> None:
        contentPath = self.__makeVersionedPath(group, key) / path
        self._deleteDirectory(contentPath)

    def getItemLargeFilesPath(
        self, group: ItemStorageGroupEnum, key: str
    ) -> Path:
        path = (
            self._unversionedFilesPath
            / group.value
            / makeStorageKey(key)
            / self.CONTENTS_PATH
        )
        path.mkdir(mode=DIR_CREATE_MODE, parents=True, exist_ok=True)
        return path

    @inlineCallbacks
    def downloadLargeFiles(self):
        semophore = DeferredSemaphore(4)
        deferreds = []
        archivesToBeDownloaded = []
        for item in self.getItems(ItemStorageGroupEnum.FileArchive):
            if item.tupleType() != FileArchiveTupleTypeEnum.LARGE.value:
                continue

            path = item.archivePath
            # A large file is already present
            if path.exists():
                continue

            # Download only if item has URI to download file from
            if not item.remoteUri:
                continue

            archivesToBeDownloaded.append(item)
            downloader = LargeFileHttpDownloader(
                item.remoteUri, item.archivePath
            )
            deferreds.append(semophore.run(downloader.download))

        deferredsResult = yield DeferredList(deferreds)
        for archive, (success, resultOrFailures) in zip(
            archivesToBeDownloaded, deferredsResult
        ):
            if success:
                logger.info(f"Downloaded large file for {archive.key}")
            else:
                logger.info(f"Failed to download file for {archive.key}")
            # TODO, Check result?


class GitObjectStorageContext(
    GitLibMixin, ObjectStorageContext, GitObjectStorageContextFileMixin
):
    def __init__(self, projectPath: Path, info: ContextProjectInfo):
        assert isinstance(projectPath, Path), "projectPath is not of type Path"
        GitLibMixin.__init__(self, projectPath)
        self.__projectInfo = info

        """ This is equivalent to an Identity Map in SqlAlchemy Session
        Holds references to all instances fetched from the storage
        ensuring that "one and only one" reference exists for an item
        For example, s1 = getItem("Step1"), s2 = getItem("Step1")
        id(s1) == id(s2) """
        self.__instanceByKeyByGroup = defaultdict(dict)

        self.__instanceByKeyByGroup[ItemStorageGroupEnum.Step] = {}
        self.__instanceByKeyByGroup[ItemStorageGroupEnum.Parameter] = {}
        self.__instanceByKeyByGroup[ItemStorageGroupEnum.FileArchive] = {}

    def load(self):
        # Run the migrations to bring the project to the latest revision
        # Migrations need to be run before the items are loaded and as a
        # result cannot use the public methods (getItem, addItem) in the
        # `upgrade` and `downgrade` functions
        runMigrationsForStorageContext(self)

        # Cache all items in memory upon load.
        # This will speed up the access of the data
        self.__loadItems(ItemStorageGroupEnum.Parameter)
        self.__loadItems(ItemStorageGroupEnum.FileArchive)
        self.__loadItems(ItemStorageGroupEnum.Step)

    @property
    def project(self) -> ContextProjectInfo:
        return self.__projectInfo

    def getRevision(self) -> str:
        metadata = self.metadata
        # If revision is missing return the first revision
        return metadata.get("revision", "16e34a42a9a1")

    def setRevision(self, revision: str):
        metadata = self.metadata
        metadata["revision"] = revision

        self.writeMetadata(metadata)
        self.commit("Wrote metadata.json")

    def getItem(
        self, group: ItemStorageGroupEnum, key: str
    ) -> Optional[StorageTuple]:
        assert group in self.__instanceByKeyByGroup, (
            "Group %s is not a valid" % group
        )
        itemsByKey = self.__instanceByKeyByGroup[group]

        if key not in itemsByKey:
            raise ItemNotFoundError(
                "Item not found, Group [%s], Key [%s]" % (group, key)
            )

        return itemsByKey[key]

    def getItems(
        self,
        group: ItemStorageGroupEnum,
    ) -> Iterator[StorageTuple]:
        assert group in self.__instanceByKeyByGroup, (
            "Group %s is not " "valid" % group
        )
        return iter(self.__instanceByKeyByGroup[group].values())

    def getItemMap(
        self, group: ItemStorageGroupEnum
    ) -> dict[str, StorageTuple]:
        assert group in self.__instanceByKeyByGroup, (
            "Group %s is not " "valid" % group
        )
        return self.__instanceByKeyByGroup[group].copy()

    def addItem(self, item: StorageTuple):
        item._bind(self)

        # Validate the item before storing
        self.validateItem(item)
        self.validateKeyDoesNotExist(group=item.storageGroup, itemKey=item.key)

        self.__writeItem(item)
        self.__storeReferenceToTuple(item)

    def validateItem(self, item):
        # TupleField Validation
        Tuple.restfulJsonDictToTupleWithValidation(
            item.tupleToRestfulJsonDict(), item.__class__
        )
        # Tuple Validation
        for validator in self.Validators():
            validator.validate(self, item)

    def expungeItem(self, item: StorageTuple):
        # noinspection PyProtectedMember
        item._unbind()

    def expireItem(self, item: StorageTuple) -> None:
        # noinspection PyProtectedMember
        item._expire()

    def reloadItem(self, item: StorageTuple) -> StorageTuple:
        freshItem = self.__readItem(item.__class__, item.key)
        self.__storeReferenceToTuple(freshItem)
        freshItem._bind(self)
        item._expire()
        return freshItem

    def mergeItem(self, item: StorageTuple) -> StorageTuple:
        try:
            existingItem = self.getItem(item.storageGroup, item.key)

            # If it's the same item we have in our store, then there is
            # nothing to do.
            if item is existingItem:
                # Rebind the item, and any member tuples it has
                item._bind(self)
                self.__writeItem(item)
                return item

        except ItemNotFoundError:
            existingItem = None

        if existingItem is None:
            raise ItemNotFoundError("Item does not yet exist in context")

        if item.tupleType() != existingItem.tupleType():
            raise TypeError("Items need to be of same type for merging")

        # Unbind the oldItem, bind the item
        self.expungeItem(existingItem)
        self.expireItem(existingItem)

        # noinspection PyProtectedMember
        item._bind(self)

        # Merge item in storage. We just need to store the item
        self.__writeItem(item)
        self.__storeReferenceToTuple(item)

        return item

    def updateItemKey(self, item: StorageTuple, newKey: str) -> None:
        if not item.storageContext:
            raise Exception(
                "Item must be merged using old key before updating to new key"
            )
        oldKey = item.key

        self.validateKeyDoesNotExist(item.storageGroup, newKey)

        # Update all the objects that reference this key
        self._cascadeUpdateKey(item, newKey)

        # Rename the item from the instance cache
        self.__removeReferenceToTuple(item)

        # Update the key
        item.key = newKey

        # Move the whole tree of contents over first, to ensure we include
        # any non-json content (EG, the versioned file archives)
        self.__moveItemDirectory(item.storageGroup, oldKey, newKey)

        # Move any unversioned storage, their storage is based on the key
        self.__moveItemUnversionedStorage(item.storageGroup, oldKey, newKey)

        # Write the updated json files, overwriting some of the files we just
        # moved
        self.__writeItem(item)

        # Store the new reference
        self.__storeReferenceToTuple(item)

    def __storeReferenceToTuple(self, item):
        assert (
            item.storageContext is self
        ), "Item must be bound before it is cached"
        self.__instanceByKeyByGroup[item.storageGroup][item.key] = item

    def __removeReferenceToTuple(self, item):
        self.__instanceByKeyByGroup[item.storageGroup].pop(item.key, None)

    def __loadItems(self, group: ItemStorageGroupEnum) -> None:
        startTime = datetime.now(pytz.utc)
        try:
            tree = self._getTree(Path(group.value))
        except FileNotFoundError:
            return

        BaseClass = GitObjectStorageContext._BaseClassByGroup[group]

        errors = []
        for object_tree in tree:
            # object_tree.name is the sanitized key (i.e. folder name) and not
            # the name of the tuple
            try:
                item = self.__readItem(BaseClass, object_tree.name)
            except (TypeError, ItemNotFoundError) as e:
                errors.append(
                    "Group %s, Key %s, Error: %s"
                    % (group.value, object_tree.name, str(e))
                )
                continue

            item._bind(self)
            self.__storeReferenceToTuple(item)

        if errors:
            raise ProjectValidationError(errors)

        logger.debug(
            "Completed loading group %s in %s",
            group,
            datetime.now(pytz.utc) - startTime,
        )

    def __writeItem(self, item: type[StorageTuple]) -> None:
        fields = item.tupleToRestfulJsonDict()
        comment = fields.pop("comment", None)
        # Scripts and SQL is stored under a `script.txt` file
        script = fields.pop("script", None) or fields.pop("sql", None)

        # FIXME If the name isn't suitable for storage
        #  then an exception needs raising instead.
        sanitizedItemKey = makeStorageKey(item.key)
        group = item.storageGroup.value
        itemPath = Path(f"{group}/{sanitizedItemKey}")

        self._writeFile(
            (itemPath / "metadata.json").as_posix(),
            json.dumps(
                fields,
                indent=4,
                sort_keys=True,
                separators=(", ", ": "),
            ).encode(),
        )

        if comment:
            self._writeFile(
                (itemPath / "comment.md").as_posix(), comment.encode()
            )
        if script:
            self._writeFile(
                (itemPath / "script.txt").as_posix(), script.encode()
            )

    def __moveItemDirectory(
        self, group: ItemStorageGroupEnum, oldKey: str, newKey: str
    ):
        oldKeyPath = Path(f"{group.value}/{oldKey}")
        newKeyPath = Path(f"{group.value}/{newKey}")
        self._moveDirectory(oldKeyPath, newKeyPath)

    def __moveItemUnversionedStorage(
        self,
        group: ItemStorageGroupEnum,
        key: str,
        newKey: str,
    ) -> None:
        oldPath = self.getItemLargeFilesPath(group, key)
        newPath = self.getItemLargeFilesPath(group, newKey)

        if oldPath.exists():
            newPath.parent.mkdir(
                mode=DIR_CREATE_MODE, parents=True, exist_ok=True
            )
            if newPath.exists():
                shutil.rmtree(newPath)

            shutil.move(oldPath, newPath.parent)

            if oldPath.parent.exists():
                shutil.rmtree(oldPath.parent)

    def __deleteItem(self, item: type[StorageTuple]) -> None:
        group = item.storageGroup.value
        itemPath = Path(f"{group}/{item.key}")
        self._deleteDirectory(itemPath)

    def __readItem(
        self, ItemClass: type[StorageTuple], key: str
    ) -> StorageTuple:
        groupStr = ItemClass.storageGroup.value
        path = Path(f"{groupStr}/{makeStorageKey(key)}")
        try:
            data = self._readFile(path / "metadata.json")
            metadataStr = data.decode()
            metadata = json.loads(metadataStr)
        except FileNotFoundError:
            raise ItemNotFoundError(f"{key} does not exist in {groupStr}")

        try:
            commentData = self._readFile(path / "comment.md")
            if commentData is None:
                # Every item has an comment associated with it. If it does not
                # exist we need to set it to None
                metadata["comment"] = None
            else:
                metadata["comment"] = commentData.decode()
        except FileNotFoundError:
            metadata["comment"] = ""

        try:
            scriptData = self._readFile(path / "script.txt")
            # Not every item has target 'script' or `sql` associated with it
            if scriptData is not None:
                metadata[
                    "sql"
                    if metadata["type"] == StepSqlOracleTuple.tupleType()
                    else "script"
                ] = scriptData.decode()
        except FileNotFoundError:
            pass

        # noinspection PyTypeChecker
        return Tuple.restfulJsonDictToTupleWithValidation(metadata, ItemClass)

    def deleteItem(self, item: StorageTuple):
        if item.storageContext is None:
            raise ValueError(f"ObjectStorageContext not bound to {item.name}")

        self.__deleteItem(item)

        self.__removeReferenceToTuple(item)
        self.expungeItem(item)
        self.expireItem(item)
        self._cascadeDelete(item)



from pathlib import Path

from attune_project_api.ObjectStorageContext import ArchiveFileContent
from attune_project_api._private.archive_format import guessArchiveFormat
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_7Z,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR_BZ2,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR_GZ,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_TAR_XZ,
)
from attune_project_api.items.file_archive_tuples.file_tuple_constants import (
    ARCHIVE_FORMAT_KEY_ZIP,
)

__ARCHIVE_LISTER_BY_FORMAT_KEY = {}

MAX_EXTRACT_SIZE = 10 * 1024 * 1024


def __registerArchiveContentGetter(key: str):
    def wrapper(getterMethod):
        __ARCHIVE_LISTER_BY_FORMAT_KEY[key] = getterMethod
        return getterMethod

    return wrapper


def getArchiveFileContent(
    archivePath: Path, memberPath: Path
) -> ArchiveFileContent:
    format = guessArchiveFormat(archivePath)
    getter = __ARCHIVE_LISTER_BY_FORMAT_KEY[format.key]
    return getter(archivePath, memberPath)


def __assureSmallSize(path: Path, size: int) -> None:
    if size > MAX_EXTRACT_SIZE:
        raise Exception(
            "Requested file is more than %s, we're not "
            "loading that into memory. %s (%s)" % (MAX_EXTRACT_SIZE, path, size)
        )


@__registerArchiveContentGetter(ARCHIVE_FORMAT_KEY_7Z)
def _get7z(archivePath: Path, memberPath: Path) -> ArchiveFileContent:
    from py7zr import SevenZipFile

    pathStr = str(memberPath)

    with SevenZipFile(archivePath, "r") as reader:
        info = next(filter(lambda o: o.filename == pathStr, reader.files))
        __assureSmallSize(memberPath, info.uncompressed)
        data = reader.read([pathStr])[pathStr].read()
        return ArchiveFileContent(memberPath, data, False)


@__registerArchiveContentGetter(ARCHIVE_FORMAT_KEY_ZIP)
def _getZip(archivePath: Path, memberPath: Path) -> ArchiveFileContent:
    from zipfile import ZipFile, ZIP_DEFLATED, ZipInfo

    pathStr = str(memberPath)

    with ZipFile(archivePath, "r", ZIP_DEFLATED) as reader:
        info: ZipInfo = reader.getinfo(pathStr)
        __assureSmallSize(memberPath, info.file_size)
        data = reader.read(pathStr)
        return ArchiveFileContent(memberPath, data, False)


@__registerArchiveContentGetter(ARCHIVE_FORMAT_KEY_TAR)
@__registerArchiveContentGetter(ARCHIVE_FORMAT_KEY_TAR_GZ)
@__registerArchiveContentGetter(ARCHIVE_FORMAT_KEY_TAR_BZ2)
@__registerArchiveContentGetter(ARCHIVE_FORMAT_KEY_TAR_XZ)
def _getTarFile(archivePath: Path, memberPath: Path) -> ArchiveFileContent:
    from tarfile import TarFile, TarInfo

    pathStr = str(memberPath)

    with TarFile.open(archivePath, "r") as reader:
        info: TarInfo = reader.getmember(pathStr)
        __assureSmallSize(memberPath, info.size)
        data = reader.extractfile(pathStr).read()
        return ArchiveFileContent(memberPath, data, False)



from .ObjectStorageContext import ObjectStorageContext
from .StorageTuple import StorageTuple
from .TupleFieldValidators import *
from .TupleValidators import *
from .tuples import *

__project__ = "Attune Project API"
__copyright__ = "2022, ServerTribe HQ Pty Ltd"
__author__ = "ServerTribe"
__version__ = '0.2.2'



class ItemNotFoundError(Exception):
    pass


class NonUniqueNameError(ValueError):
    pass


class NonUniqueScriptRefError(ValueError):
    pass


class InvalidReferenceError(ValueError):
    pass


class NoRemoteDefined(ValueError):
    pass


class MergeConflict(Exception):
    pass


class NoChangesToCommitError(Exception):
    pass


class ChangesNotCommitedError(Exception):
    pass


class ProjectValidationError(Exception):
    def __init__(self, errors: list[str]):
        Exception.__init__(
            self, "%s errors occurred during load of project" % len(errors)
        )
        self.errors: list[str] = errors



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""

from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from . import addStepDeclarative
from . import extractTextPlaceholders
from .step_tuple import StepTupleTypeEnum
from .. import NotZeroLenStr
from ... import ParameterTuple
from ... import StepTuple
from ...ObjectStorageContext import ObjectStorageContext
from ...RelationField import RelationField


SSH_STEP_INTERPRETERS = {}


class StepSshInterpreter:
    def __init__(self, _id, name, command, head, tail, uiAceMode):
        self.id, self.name, self.command = _id, name, command
        self.head, self.tail = head, tail
        self.uiAceMode = uiAceMode
        SSH_STEP_INTERPRETERS[self.id] = self


stepSshIntBash = StepSshInterpreter(
    1, "bash", "bash -l", "set -o nounset; set -o errexit;", "", "sh"
)

stepSshIntPython = StepSshInterpreter(
    2, "python", "python -u", "", "", "python"
)

stepSshIntPerl = StepSshInterpreter(3, "perl", "perl", "", "exit 0;", "perl")


@ObjectStorageContext.registerItemClass
@addStepDeclarative("Execute Linux Script")
@addTupleType
class StepSshTuple(StepTuple):
    __tupleType__ = StepTupleTypeEnum.SSH.value

    script: NotZeroLenStr = TupleField()
    interpreter: int = TupleField(defaultValue=stepSshIntBash.id)
    serverKey: NotZeroLenStr = TupleField()
    osCredKey: NotZeroLenStr = TupleField()
    successExitCode: int = TupleField(defaultValue=0)

    server: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="serverKey",
    )
    osCred: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="osCredKey",
    )

    def parameters(self) -> list["ParameterTuple"]:
        return [self.server, self.osCred]

    def scriptReferences(self) -> list[str]:
        return extractTextPlaceholders(self.script)



import os
import shutil

from setuptools import find_packages
from setuptools import setup

pip_package_name = "attune-project-api"
py_package_folder = "attune_project_api"
package_version = '0.2.2'

egg_info = "%s.egg-info" % pip_package_name
if os.path.isdir(egg_info):
    shutil.rmtree(egg_info)

if os.path.isfile("MANIFEST"):
    os.remove("MANIFEST")

includePathContains = ("alembic_migrations", "alembic.ini")
excludePathContains = ("__pycache__", "platforms", "dist")
includeFilesStartWith = ()
excludeFilesEndWith = (".pyc", ".Apple")
excludeFilesStartWith = ("test", "tests")


def find_package_files():
    paths = []
    for (path, directories, filenames) in os.walk(py_package_folder):
        if not [e for e in includePathContains if e in path]:
            if [e for e in excludePathContains if e in path]:
                continue

        for filename in filenames:
            if not [e for e in includeFilesStartWith if filename.startswith(e)]:
                if [e for e in excludeFilesEndWith if filename.endswith(e)]:
                    continue

                if [e for e in excludeFilesStartWith if filename.startswith(e)]:
                    continue

            relPath = os.path.join(path, filename)
            paths.append(relPath[len(py_package_folder) + 1 :])

    return paths


package_files = find_package_files()

requirements = [
    "twisted",
    "pygit2",
    "vortexpy==2.6.*",
    "pytz",
    "pathvalidate",
    # Support for inspecting 7z archives
    "py7zr",
    "markdown",
    "alembic",
]


doc_requirements = [
    "sphinx",
    "sphinx-rtd-theme",
    "sphinx-autobuild",
    "pytmpdir",
]

requirements.extend(doc_requirements)

setup(
    name=pip_package_name,
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"]
    ),
    package_data={"": package_files},
    install_requires=requirements,
    zip_safe=False,
    version=package_version,
    description="",
    author="ServerTribe",
    author_email="support@servertribe.com",
    classifiers=["Programming Language :: Python :: 3.9"],
)



"""
*
 *  Copyright ServerTribe HQ Pty Ltd 2021
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  ServerTribe HQ Pty Ltd
 *
"""
from pathlib import Path
from typing import Optional
from typing import Union

from vortex.Tuple import PolymorphicTupleTypeFieldArg
from vortex.Tuple import TupleField

from attune_project_api import ObjectStorageContext
from attune_project_api.ObjectStorageContext import ArchiveFileContent
from attune_project_api.ObjectStorageContext import ArchiveFileInfo
from attune_project_api.ObjectStorageContext import VersionedFileContent
from attune_project_api.ObjectStorageContext import VersionedFileInfo
from attune_project_api.StorageTuple import ItemStorageGroupEnum
from attune_project_api.StorageTuple import StorageTuple


@ObjectStorageContext.registerItemClass
class FileArchiveTuple(StorageTuple):
    """File Archive Tuple

    This is the base archive tuple.
    An archive (displayed as "Files" in the UI) is just that, an archive

    """

    __tupleArgs__ = (PolymorphicTupleTypeFieldArg("type"),)
    __group__ = ItemStorageGroupEnum.FileArchive

    key: str = TupleField()
    name: str = TupleField()
    type: str = TupleField()
    comment: Optional[str] = TupleField()

    #: This is for the user interface only
    uiData: Optional[dict] = TupleField({}, jsonExclude=True)

    @classmethod
    @property
    def niceName(cls) -> str:
        # This is meant to be an abstract method
        raise NotImplementedError()

    @property
    def listFiles(self) -> list[Union[ArchiveFileInfo, VersionedFileInfo]]:
        # This is meant to be an abstract method
        raise NotImplementedError()

    @property
    def getFileContent(
        self, path: Path
    ) -> Union[list[ArchiveFileContent], list[VersionedFileContent]]:
        # This is meant to be an abstract method
        raise NotImplementedError()

    def _makeFileName(self, formatKey: str) -> str:
        allowedLetters = {chr(i) for i in range(0x30, 0x39)}  # 0-9
        allowedLetters |= {chr(i) for i in range(0x41, 0x5A)}  # A-Z
        allowedLetters |= {chr(i) for i in range(0x61, 0x7A)}  # a-z
        allowedLetters.add("_")

        name = self.name.replace(" ", "_")
        name = "".join([c for c in name if c in allowedLetters])

        return f"{name}.{formatKey}"



from attune_project_api import ParameterTuple
from attune_project_api.RelationField import RelationField
from attune_project_api._contexts import GitObjectStorageContext
from attune_project_api.items import NotZeroLenStr
from attune_project_api.items.file_archive_tuples.file_archive_tuple import (
    FileArchiveTuple,
)
from attune_project_api.items.step_tuples import addStepDeclarative
from attune_project_api.items.step_tuples import extractTextPlaceholders
from attune_project_api.items.step_tuples.step_tuple import StepTuple
from attune_project_api.items.step_tuples.step_tuple import StepTupleTypeEnum
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType


@GitObjectStorageContext.registerItemClass
@addStepDeclarative("Push Files")
@addTupleType
class StepPushDesignFileTuple(StepTuple):
    __tupleType__ = StepTupleTypeEnum.PUSH_DESIGN_FILE.value

    serverKey: NotZeroLenStr = TupleField()
    osCredKey: NotZeroLenStr = TupleField()
    deployPath: NotZeroLenStr = TupleField()
    archiveKey: NotZeroLenStr = TupleField()
    unpack: bool = TupleField(False)

    server: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="serverKey",
    )
    osCred: ParameterTuple = RelationField(
        ForeignClass=ParameterTuple,
        referenceKeyFieldName="osCredKey",
    )
    archive: FileArchiveTuple = RelationField(
        ForeignClass=FileArchiveTuple,
        referenceKeyFieldName="archiveKey",
    )

    def parameters(self) -> list["ParameterTuple"]:
        return [self.server, self.osCred]

    def scriptReferences(self) -> list[str]:
        return extractTextPlaceholders(self.deployPath)



from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from attune_project_api import ObjectStorageContext
from attune_project_api import ParameterTuple
from attune_project_api.items import NotZeroLenStr
from attune_project_api.items.step_tuples import addStepDeclarative
from attune_project_api.items.step_tuples import extractTextPlaceholders
from attune_project_api.items.step_tuples.step_ssh_tuple import StepSshTuple
from attune_project_api.items.step_tuples.step_tuple import StepTupleTypeEnum


@ObjectStorageContext.registerItemClass
@addStepDeclarative("Execute Linux Script With Responses")
@addTupleType
class StepSshPromptedTuple(StepSshTuple):
    __tupleType__ = StepTupleTypeEnum.SSH_PROMPTED.value

    promptResponses: NotZeroLenStr = TupleField()
    type: NotZeroLenStr = TupleField()

    def parameters(self) -> list["ParameterTuple"]:
        return [self.server, self.osCred]

    def scriptReferences(self) -> list[str]:
        return extractTextPlaceholders(self.script + self.promptResponses)
